'******************************************************************************
'*  SrdCtrls.bi                                                               *
'*  Copyright  2006-2007 Nastase Eodor                                       *
'*  This file is part of Paradigma GUI Library                                *
'******************************************************************************

'Option Explicit

Const BS_MASK = &H000F

Enum TMessageType
     mtWarning      = 0
     mtError        = 1
     mtInformation  = 2
     mtConfirmation = 3
     mtCustom       = 4
End Enum
Enum TMessageButton
     mbOk               = 1
     mbOkCancel         = 2
     mbYesNo            = 3
     mbYesNoCancel      = 4
     mbRetryCancel      = 5
     mbAbortRetryIgnore = 6
End Enum

Declare Function InputBox(byval Prompt As String,byval Caption As String,byval Text As String = "") As String
Declare Function MessageDlg (byval Message As String,byval MsgType As TMessageType,byval MsgButton As Integer,byval HelpContext As Long) As Integer
Declare Sub ShowMessage(byval Message As String)

'################
'#    TLabel    #
'################

Enum TStaticBorderStyle
     stNone    = 0
     stSimple  = SS_SIMPLE
     stSunken  = SS_SUNKEN
End Enum
Enum TStaticAlignment
     staLeft   = SS_LEFT
     staCenter = SS_CENTER
     staRight  = SS_RIGHT
End Enum     
Type TLabel
     Align        As TAlign
     Alignment    As TStaticAlignment
     BorderStyle  As TStaticBorderStyle
     Caption      As String
     ClientRect   As RECT
     Color        As Integer
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     HelpContext  As Integer
     Hint         As String
     Left         As Integer
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     ShowHint     As Boolean
     Tag          As Integer
     Top          As Integer
     Visible      As Boolean
     Width        As Integer
     
     BringToFront As Sub(Self As TLabel)
     Create       As Function(Self As TLabel) As HWND
     Free         As Sub(Self As TLabel)
     
     SetAlign       As Sub(Self As TLabel,byval fAlign As TAlign)
     SetAlignment   As Sub(Self As TLabel,byval fAlignment As TStaticAlignment)
     SetBorderStyle As Sub(Self As TLabel,byval fBorderStyle As TStaticBorderStyle)
     SetCaption     As Sub(Self As TLabel,byval fCaption As String)
     SetColor       As Sub(Self As TLabel,byval fColor As Integer)
     SetCursor      As Sub(Self As TLabel,byval fCursor As HCURSOR)
     SetEnabled     As Sub(Self As TLabel,byval fEnabled As Boolean)
     SetFont        As Sub(Self As TLabel,fFont As TFont)
     SetHeight      As Sub(Self As TLabel,byval fHeight As Integer)
     SetHint        As Sub(Self As TLabel,byval fHint As String)
     SetLeft        As Sub(Self As TLabel,byval fLeft As Integer)
     SetParent      As Sub(Self As TLabel,byval fParent As HWND)
     SetPopupMenu   As Sub(Self As TLabel,PopupMenu As TPopupMenu)
     SetShowHint    As Sub(Self As TLabel,byval fShowHint As Boolean)
     SetTop         As Sub(Self As TLabel,byval fTop As Integer)
     SetVisible     As Sub(Self As TLabel,byval fVisible As Boolean)
     SetWidth       As Sub(Self As TLabel,byval fWidth As Integer)
    
     OnClick      As Sub(Self As TLabel)
     OnDblClick   As Sub(Self As TLabel)
     OnKeyDown    As Sub(Self As TLabel,Key As Word)
     OnKeyPress   As Sub(Self As TLabel,Key As Byte) 
     OnKeyUp      As Sub(Self As TLabel,Key As Word)
     OnMouseDown  As Sub(Self As TLabel,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
     OnMouseMove  As Sub(Self As TLabel,byval X As Integer,byval Y As Integer,byval Shift As Integer)
     OnMouseUp    As Sub(Self As TLabel,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
End Type

Dim Shared TLabel_DoClick      As Sub(Self As TLabel)
Dim Shared TLabel_DoDblClick   As Sub(Self As TLabel)
Dim Shared TLabel_DoKeyDown    As Sub(Self As TLabel,Key As Word)
Dim Shared TLabel_DoKeyPress   As Sub(Self As TLabel,Key As Byte)
Dim Shared TLabel_DoKeyUp      As Sub(Self As TLabel,Key As Word)
Dim Shared TLabel_DoMouseDown  As Sub(Self As TLabel,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
Dim Shared TLabel_DoMouseMove  As Sub(Self As TLabel,byval X As Integer,byval Y As Integer,byval Shift As Integer)
Dim Shared TLabel_DoMouseUp    As Sub(Self As TLabel,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
     
Declare Function TLabel_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TLabel_Create(Self As TLabel) As HWND
Declare Sub TLabel_Free(Self As TLabel)
Declare Sub TLabel_BringToFront(Self As TLabel)
Declare Sub TLabel_Initialize(Self As TLabel)
Declare Sub TLabel_SetAlign(Self As TLabel,byval fAlign As TAlign)
Declare Sub TLabel_SetAlignment(Self As TLabel,byval fAlignment As TStaticAlignment)
Declare Sub TLabel_SetAutosize(Self As TLabel,byval fAutoSize As Boolean)
Declare Sub TLabel_SetBorderStyle(Self As TLabel,byval fBorderStyle As TStaticBorderStyle)
Declare Sub TLabel_SetCaption(Self As TLabel,byval fCaption As String)
Declare Sub TLabel_SetHint(Self As TLabel,byval fHint As String)
Declare Sub TLabel_SetShowHint(Self As TLabel,byval fShowHint As Boolean)
Declare Sub TLabel_SetCursor(Self As TLabel,byval fCursor As HCURSOR)
Declare Sub TLabel_SetColor(Self As TLabel,byval fColor As Integer)
Declare Sub TLabel_SetLeft(Self As TLabel,byval fLeft As Integer)
Declare Sub TLabel_SetTop(Self As TLabel,byval fTop As Integer)
Declare Sub TLabel_SetWidth(Self As TLabel,byval fWidth As Integer)
Declare Sub TLabel_SetHeight(Self As TLabel,byval fHeight As Integer)
Declare Sub TLabel_SetVisible(Self As TLabel,byval fVisible As Boolean)
Declare Sub TLabel_SetParent(Self As TLabel,byval fParent As HWND)
Declare Sub TLabel_SetEnabled(Self As TLabel,byval fEnabled As Boolean)
Declare Sub TLabel_SetFont(Self As TLabel,fFont As TFont)
Declare Sub TLabel_SetPopupMenu(Self As TLabel,fPopupMenu As TPopupMenu)

Sub TLabel_Initialize(Self As TLabel)
    Self.Create          = @TLabel_Create
    Self.BringToFront    = @TLabel_BringToFront
    Self.Free            = @TLabel_Free
    Self.SetAlign        = @TLabel_SetAlign
    Self.SetAlignment    = @TLabel_SetAlignment
    Self.SetBorderStyle  = @TLabel_SetBorderStyle
    Self.SetCaption      = @TLabel_SetCaption
    Self.SetHint         = @TLabel_SetHint
    Self.SetShowHint     = @TLabel_SetShowHint
    Self.SetColor        = @TLabel_SetColor
    Self.SetCursor       = @TLabel_SetCursor
    Self.SetLeft         = @TLabel_SetLeft
    Self.SetTop          = @TLabel_SetTop
    Self.SetWidth        = @TLabel_SetWidth
    Self.SetHeight       = @TLabel_SetHeight
    Self.SetEnabled      = @TLabel_SetEnabled
    Self.SetVisible      = @TLabel_SetVisible
    Self.SetParent       = @TLabel_SetParent
    Self.SetFont         = @TLabel_SetFont
    Self.SetPopupMenu    = @TLabel_SetPopupMenu
    Self.Create(Self)
End Sub

Function TLabel_Create(Self As TLabel) As HWND
    If Self.Alignment = 0 then Self.Alignment = staLeft
    Self.Params.Style     = SS_NOTIFY OR Self.BorderStyle OR Self.Alignment
    Self.Params.TypeName = "TLabel"
    If Self.Width  = 0 then Self.Width  = 75
    If Self.Height = 0 then Self.Height = 17
    
    Self.Handle = CreateWindow("Static",Self.Caption,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    
    Self.Params.HintHandle = CreateWindowEx(0,TOOLTIPS_CLASS,"",TTS_ALWAYSTIP, 0,0,0,0,0,NULL,Instance,NULL)

    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SUBCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)

    SendMessage(Self.Params.HintHandle,TTM_ADDTOOL, 0,cint(@Self.Params.Ti))
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)

    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TLabel_Proc)))
    
    If Self.Color = &H0 then Self.Color = GetSysColor(COLOR_BTNFACE)
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    SetProp(Self.Handle,"HelpContext",cast(Integer Ptr,Self.HelpContext))
    
    Application.Controls.Add(Application.Controls,Self.Handle)
    
    Return Self.Handle
End Function

Sub TLabel_SetAlign(Self As TLabel,byval fAlign As TAlign)
    Self.Align = fAlign
End Sub

Sub TLabel_SetAlignment(Self As TLabel,byval fAlignment As TStaticAlignment)
    Self.Alignment = fAlignment
    Self.Free(Self)
    Self.Create(Self)
End Sub

Sub TLabel_SetBorderStyle(Self As TLabel,byval fBorderStyle As TStaticBorderStyle)
    Self.BorderStyle = fBorderStyle
    Self.Free(Self)
    Self.Create(Self)
End Sub

Sub TLabel_Free(Self As TLabel)
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If   
End Sub

Sub TLabel_BringToFront(Self As TLabel)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TLabel_SetCursor(Self As TLabel,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TLabel_SetColor(Self As TLabel,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    InvalidateRect(Self.Handle,0,True)
End Sub

Sub TLabel_SetHint(Self As TLabel,byval fHint As String)
    Self.Hint = fHint
    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SubCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)
    SendMessage(Self.Params.HintHandle, TTM_UPDATETIPTEXT, 0,cint(@Self.Params.Ti))
End Sub

Sub TLabel_SetShowHint(Self As TLabel,byval fShowHint As Boolean)
    Self.ShowHint = fShowHint
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)
End Sub

Sub TLabel_SetCaption(Self As TLabel,byval fCaption As String)
    Self.Caption = fCaption
    If Self.Handle then SetWindowText(Self.Handle,Self.Caption)
End Sub

Sub TLabel_SetLeft(Self As TLabel,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TLabel_SetTop(Self As TLabel,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then  MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TLabel_SetWidth(Self As TLabel,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TLabel_SetHeight(Self As TLabel,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TLabel_SetEnabled(Self As TLabel,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TLabel_SetVisible(Self As TLabel,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TLabel_SetFont(Self As TLabel,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If   
End Sub

Sub TLabel_SetPopupMenu(Self As TLabel,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TLabel_SetParent(Self As TLabel,byval fParent As HWND)
    If fParent <> Self.Parent then
       Self.Parent = fParent
       If Self.Handle then SetParent(Self.Handle,Self.Parent)
    End If
End Sub

Function TLabel_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TLabel Ptr
    Dim As RECT WR   
    Dim As POINT PT
    Dim As TSmallPoint SP
    Dim As Integer X,Y,CommandID,I,ID
    Ctrl = Cast(TLabel Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
           Case WM_RBUTTONUP  :X = LoWord(lParam)
                                Y = HiWord(lParam)
                                PT.X = X
                                PT.Y = Y
                                If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
           Case WM_LBUTTONUP  : X = HiWord(lParam)
                                 Y = LoWord(lParam)
                                 If Ctrl->OnClick <> 0 then
                                    TLabel_DoClick = Ctrl->OnClick
                                    TLabel_DoClick(*Ctrl)
                                 End If
                                 If Ctrl->OnMouseUp <> 0 then
                                    TLabel_DoMouseUp = Ctrl->OnMouseUp
                                    Ctrl->OnMouseUp(*Ctrl,0,X,Y,wParam And &HFFFF)
                                 End If  
           Case WM_LBUTTONDOWN: X = HiWord(lParam)
                                Y = LoWord(lParam)
                                If Ctrl->OnMouseDown <> 0 then
                                   TLabel_DoMouseDown = Ctrl->OnMouseDown
                                   Ctrl->OnMouseDown(*Ctrl,0,X,Y,wParam And &HFFFF)
                                End If
           Case WM_MOUSEMOVE  : If Ctrl->OnMouseMove <> 0 then
                                   TLabel_DoMouseMove = Ctrl->OnMouseMove
                                   Ctrl->OnMouseMove(*Ctrl,X,Y,wParam And &HFFFF)
                                End If                     
           Case WM_SIZE       : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                GetWindowRect(Ctrl->Handle,@WR)
                                Ctrl->Width  = WR.Right - WR.Left
                                Ctrl->Height = WR.Bottom - WR.Top
           Case WM_SIZING     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                GetWindowRect(Ctrl->Handle,@WR)
                                Ctrl->Width  = WR.Right - WR.Left
                                Ctrl->Height = WR.Bottom - WR.Top
           Case WM_MOVE       : GetWindowRect(Ctrl->Handle,@WR)
                                PT.X = WR.Left
                                PT.Y = WR.Top
                                ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                Ctrl->Left = PT.X
                                Ctrl->Top  = PT.Y
          Case WM_MOVING     :  GetWindowRect(Ctrl->Handle,@WR)
                                 PT.X = WR.Left
                                 PT.Y = WR.Top
                                 ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                 Ctrl->Left = PT.X
                                 Ctrl->Top  = PT.Y
          Case WM_MOUSEMOVE  : 
          Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If                        
          Case WM_DESTROY    : RemoveControlProps(Ctrl->Handle)
          Case WM_COMMAND    : If HiWord(wParam) = 0 then 'is itemmenu
                                  CommandID = LoWord(wParam)
                                  If Ctrl->PopupMenu.Handle then
                                     If Ctrl->PopupMenu.Count = 0 then Exit Function 
                                         For i = 0 to Ctrl->PopupMenu.Count -1 
                                             If Ctrl->PopupMenu.Items[i]->CommandID = CommandID then
                                                Id = i 
                                                Exit For
                                             End If
                                         Next i 
                                         If Ctrl->PopupMenu.Items[Id]->OnClick <> 0 then
                                             TMenuItem_DoClick = Ctrl->PopupMenu.Items[Id]->OnClick
                                             TMenuItem_DoClick(*Ctrl->PopupMenu.Items[Id])
                                         End If
                                      End If
                                   End If
   End Select
   If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
   Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#   TButton    #
'################

Type TButton
     Align        As TAlign
     Caption      As String
     ClientRect   As RECT
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     HelpContext  As Integer
     Hint         As String
     Left         As Integer
     ModalResult  As TModalResult
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     ShowHint     As Boolean
     Tag          As Integer
     Top          As Integer
     Visible      As Boolean
     Width        As Integer
     
     Create       As Function(Self As TButton) As HWND
     BringToFront As Sub(Self As TButton)
     Free         As Sub(Self As TButton)
     SetFocus     As Sub(Self As TButton)
     
     SetAlign     As Sub(Self As TButton,byval fAlign As TAlign)
     SetCaption   As Sub(Self As TButton,byval fCaption As String)
     SetCursor    As Sub(Self As TButton,byval fCursor As HCURSOR)
     SetEnabled   As Sub(Self As TButton,byval fEnabled As Boolean)
     SetFont      As Sub(Self As TButton,fFont As TFont)
     SetHeight    As Sub(Self As TButton,byval fHeight As Integer)
     SetHint      As Sub(Self As TButton,byval fHint As String)
     SetLeft      As Sub(Self As TButton,byval fLeft As Integer)
     SetParent    As Sub(Self As TButton,byval fParent As HWND)
     SetPopupMenu As Sub(Self As TButton,PopupMenu As TPopupMenu)
     SetShowHint  As Sub(Self As TButton,byval fShowHint As Boolean)
     SetTop       As Sub(Self As TButton,byval fTop As Integer)
     SetVisible   As Sub(Self As TButton,byval fVisible As Boolean)
     SetWidth     As Sub(Self As TButton,byval fWidth As Integer)
    
     OnClick      As Sub(Self As TButton)
     OnKeyDown    As Sub(Self As TButton,Key As Word)
     OnKeyPress   As Sub(Self As TButton,Key As Byte) 
     OnKeyUp      As Sub(Self As TButton,Key As Word)
     OnMouseDown  As Sub(Self As TButton,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
     OnMouseMove  As Sub(Self As TButton,byval X As Integer,byval Y As Integer,byval Shift As Integer)
     OnMouseUp    As Sub(Self As TButton,byval Button As Integer,byval X As Integer,byval Y As Integer,byval Shift As Integer)
End Type

Dim Shared TButton_DoClick    As Sub(Self As TButton)

Declare Function TButton_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TButton_Create(Self As TButton) As HWND
Declare Sub TButton_Free(Self As TButton)
Declare Sub TButton_SetFocus(Self As TButton)
Declare Sub TButton_BringToFront(Self As TButton)
Declare Sub TButton_Initialize(Self As TButton)
Declare Sub TButton_SetAlign(Self As TButton,byval fAlign As TAlign)
Declare Sub TButton_SetCaption(Self As TButton,byval fCaption As String)
Declare Sub TButton_SetHint(Self As TButton,byval fHint As String)
Declare Sub TButton_SetShowHint(Self As TButton,byval fShowHint As Boolean)
Declare Sub TButton_SetCursor(Self As TButton,byval fCursor As HCURSOR)
Declare Sub TButton_SetLeft(Self As TButton,byval fLeft As Integer)
Declare Sub TButton_SetTop(Self As TButton,byval fTop As Integer)
Declare Sub TButton_SetWidth(Self As TButton,byval fWidth As Integer)
Declare Sub TButton_SetHeight(Self As TButton,byval fHeight As Integer)
Declare Sub TButton_SetVisible(Self As TButton,byval fVisible As Boolean)
Declare Sub TButton_SetParent(Self As TButton,byval fParent As HWND)
Declare Sub TButton_SetEnabled(Self As TButton,byval fEnabled As Boolean)
Declare Sub TButton_SetFont(Self As TButton,fFont As TFont)
Declare Sub TButton_SetPopupMenu(Self As TButton,fPopupMenu As TPopupMenu)

Sub TButton_Initialize(Self As TButton)
    Self.Create       = @TButton_Create
    Self.BringToFront = @TButton_BringToFront
    Self.Free         = @TButton_Free
    Self.SetFocus     = @TButton_SetFocus
    Self.SetAlign     = @TButton_SetAlign
    Self.SetCaption   = @TButton_SetCaption
    Self.SetHint      = @TButton_SetHint
    Self.SetShowHint  = @TButton_SetShowHint
    Self.SetCursor    = @TButton_SetCursor
    Self.SetLeft      = @TButton_SetLeft
    Self.SetTop       = @TButton_SetTop
    Self.SetWidth     = @TButton_SetWidth
    Self.SetHeight    = @TButton_SetHeight
    Self.SetEnabled   = @TButton_SetEnabled
    Self.SetVisible   = @TButton_SetVisible
    Self.SetParent    = @TButton_SetParent
    Self.SetFont      = @TButton_SetFont
    Self.SetPopupMenu = @TButton_SetPopupMenu
    Self.Create(Self)
End Sub

Function TButton_Create(Self As TButton) As HWND
    Self.Params.Style     =  WS_TABSTOP OR BS_TEXT OR BS_DEFPUSHBUTTON OR BS_CENTER OR BS_VCENTER
    Self.Params.TypeName = "TButton"

    If Self.Width  = 0 then Self.Width  = 75
    If Self.Height = 0 then Self.Height = 25
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"Button",Self.Caption,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    SetProp(Self.Handle,"HelpContext",cast(Integer Ptr,Self.HelpContext))
    
    Self.Params.HintHandle = CreateWindowEx(0,TOOLTIPS_CLASS,"",TTS_ALWAYSTIP, 0,0,0,0,0,NULL,Instance,NULL)

    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SubCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)

    SendMessage(Self.Params.HintHandle,TTM_ADDTOOL, 0,cint(@Self.Params.Ti))
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)

    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TButton_Proc)))
    
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    Application.Controls.Add(Application.Controls,Self.Handle)
    
    Return Self.Handle
End Function

Sub TButton_SetAlign(Self As TButton,byval fAlign As TAlign)
    Self.Align = fAlign
    SetProp(Self.Handle,"Align",@Self.Align)
    Select Case Self.Align'''future implementation
           Case alNone   : 
           Case alLeft   : 
           Case alRight  : 
           Case alTop    : 
           Case alBottom : 
           Case alClient : 
    End Select
End Sub

Sub TButton_Free(Self As TButton)
    If Self.Handle then 
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TButton_SetFocus(Self As TButton)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TButton_BringToFront(Self As TButton)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TButton_SetButtonStyle(byval Handle As HWND,byval Default As Boolean)
   Dim As Word Style
   If Handle then
      If Default then Style = BS_DEFPUSHBUTTON Else Style = BS_PUSHBUTTON
      If GetWindowLong(Handle, GWL_STYLE) and BS_MASK <> Style then SendMessage(Handle, BM_SETSTYLE, Style, 1)
   End If
End Sub

Sub TButton_SetCursor(Self As TButton,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TButton_SetHint(Self As TButton,byval fHint As String)
    Self.Hint = fHint
    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SubCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)
    SendMessage(Self.Params.HintHandle, TTM_UPDATETIPTEXT, 0,cint(@Self.Params.Ti))
End Sub

Sub TButton_SetShowHint(Self As TButton,byval fShowHint As Boolean)
    Self.ShowHint = fShowHint
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)
End Sub

Sub TButton_SetCaption(Self As TButton,byval fCaption As String)
    Self.Caption = fCaption
    If Self.Handle then SetWindowText(Self.Handle,Self.Caption)
End Sub

Sub TButton_SetLeft(Self As TButton,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TButton_SetTop(Self As TButton,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TButton_SetWidth(Self As TButton,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TButton_SetHeight(Self As TButton,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TButton_SetEnabled(Self As TButton,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TButton_SetVisible(Self As TButton,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TButton_SetFont(Self As TButton,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If   
End Sub

Sub TButton_SetPopupMenu(Self As TButton,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TButton_SetParent(Self As TButton,byval fParent As HWND)
    If fParent <> Self.Parent then
       Self.Parent = fParent
       If Self.Handle then SetParent(Self.Handle,Self.Parent)
    End If
End Sub

Function TButton_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TButton Ptr
    Dim Owner As TForm Ptr
    Dim As RECT WR   
    Dim As POINT PT 
    Dim As TSmallPoint SP
    Dim As Integer X,Y,CommandID,I,Id
    Dim As HWND hOwner
    Ctrl = Cast(TButton Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
           Case WM_RBUTTONUP   : X = LoWord(lParam)
                                 Y = HiWord(lParam)
                                 PT.X = X
                                 PT.Y = Y
                                 If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
           Case WM_LBUTTONDOWN : 
           Case WM_LBUTTONUP   : SendMessage(Ctrl->Parent,CM_MODALRESULT,Ctrl->ModalResult,0)
                                  If Ctrl->ModalResult <> mrNone then
                                     hOwner = GetAncestor(Ctrl->Handle,GA_ROOTOWNER)
                                     If Ctrl->OnClick = 0 then SendMessage(hOwner,CM_CLOSE,0,0)
                                  End If
                                  If Ctrl->OnClick <> 0 then
                                    TButton_DoClick = Ctrl->OnClick
                                    TButton_DoClick(*Ctrl)
                                  End If
           Case WM_KEYDOWN     : hOwner = GetAncestor(Ctrl->Handle,GA_ROOTOWNER)
                                  Owner = Cast(TForm Ptr,GetWindowLong(hOwner,GWL_USERDATA))   
                                  If Owner->KeyPreview then SendMessage(hOwner,WM_KEYDOWN,wParam,lParam) 
           Case WM_SIZE        : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                  GetWindowRect(Ctrl->Handle,@WR)
                                  Ctrl->Width  = WR.Right - WR.Left
                                  Ctrl->Height = WR.Bottom - WR.Top
           Case WM_SIZING      : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                  GetWindowRect(Ctrl->Handle,@WR)
                                  Ctrl->Width  = WR.Right - WR.Left
                                  Ctrl->Height = WR.Bottom - WR.Top
           Case WM_MOVE        : GetWindowRect(Ctrl->Handle,@WR)
                                  PT.X = WR.Left
                                  PT.Y = WR.Top
                                  ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                  Ctrl->Left = PT.X
                                  Ctrl->Top  = PT.Y
          Case WM_MOVING       :  GetWindowRect(Ctrl->Handle,@WR)
                                  PT.X = WR.Left
                                  PT.Y = WR.Top
                                  ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                  Ctrl->Left = PT.X
                                  Ctrl->Top  = PT.Y
          Case WM_MOUSEMOVE    :  X = LoWord(lParam)
                                   Y = HiWord(lParam)
                                   PT.X = X
                                   PT.Y = Y 
          Case wM_SETCURSOR   : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                    GetCursorPos(@PT)
                                    SP = PointToSmallPoint(PT)
                                    If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                       If GetProp(Ctrl->Handle,"Cursor") then 
                                          SetCursor(Ctrl->Cursor)
                                          Exit Function
                                       End If   
                                    Else
                                       SetCursor(crArrow)
                                       Exit Function
                                    End If
                                 End If
          Case WM_DESTROY     : RemoveControlProps(Ctrl->Handle)
          Case WM_COMMAND     : If HiWord(wParam) = 0 then 'is itemmenu
                                    CommandID = LoWord(wParam)
                                    If Ctrl->PopupMenu.Handle then
                                       If Ctrl->PopupMenu.Count = 0 then Exit Function 
                                          For i = 0 to Ctrl->PopupMenu.Count -1 
                                             If Ctrl->PopupMenu.Items[i]->CommandID = CommandID then
                                                Id = i 
                                                Exit For
                                             End If
                                          Next i 
                                          If Ctrl->PopupMenu.Items[Id]->OnClick <> 0 then
                                             TMenuItem_DoClick = Ctrl->PopupMenu.Items[Id]->OnClick
                                             TMenuItem_DoClick(*Ctrl->PopupMenu.Items[Id])
                                          End If
                                       End If
                                    End If
                   
   End Select
   If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
   Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#   TCheckBox  #
'################

Type TCheckBox
     Align        As TAlign
     Caption      As String
     Checked      As Boolean
     ClientRect   As RECT
     Color        As Integer
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     Left         As Integer
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     Top          As Integer
     Visible      As Boolean 
     Width        As Integer
  
     Create       As Function(Self As TCheckBox) As HWND
     Free         As Sub(Self As TCheckBox)
     BringToFront As Sub(Self As TCheckBox)
     SetFocus     As Sub(Self As TCheckBox)
     
     SetAlign     As Sub(Self As TCheckBox,byval fAlign As TAlign)
     SetCaption   As Sub(Self As TCheckBox,byval fCaption As String)
     SetChecked   As Sub(Self As TCheckBox,byval fChecked As Boolean)
     SetColor     As Sub(Self As TCheckBox,byval fColor As Integer)
     SetCursor    As Sub(Self As TCheckBox,byval fCursor As HCURSOR)
     SetEnabled   As Sub(Self As TCheckBox,byval fEnabled As Boolean)
     SetFont      As Sub(Self As TCheckBox,Font As TFont)
     SetHeight    As Sub(Self As TCheckBox,byval fHeight As Integer)
     SetLeft      As Sub(Self As TCheckBox,byval fLeft As Integer)
     SetParent    As Sub(Self As TCheckBox,byval fParent As HWND)
     SetPopupMenu As Sub(Self As TCheckBox,PopupMenu As TPopupMenu)
     SetTop       As Sub(Self As TCheckBox,byval fTop As Integer)
     SetVisible   As Sub(Self As TCheckBox,byval fVisible As Boolean)
     SetWidth     As Sub(Self As TCheckBox,byval fWidth As Integer)
     
     OnClick      As Sub(Self As TCheckBox)
End Type

Dim Shared TCheckBox_DoClick As Sub(Self As TCheckBox)

Declare Function TCheckBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TCheckBox_Create(Self As TCheckBox) As HWND
Declare Sub TCheckBox_Free(Self As TCheckBox)
Declare Sub TCheckBox_SetFocus(Self As TCheckBox)
Declare Sub TCheckBox_BringToFront(Self As TCheckBox)
Declare Sub TCheckBox_Initialize(Self As TCheckBox)
Declare Sub TCheckBox_SetChecked(Self As TCheckBox,byval fChecked As Boolean)
Declare Sub TCheckBox_SetColor(Self As TCheckBox,byval fColor As Integer)
Declare Sub TCheckBox_SetCursor(Self As TCheckBox,byval fCursor As HCURSOR)
Declare Sub TCheckBox_SetCaption(Self As TCheckBox,byval fCaption As String)
Declare Sub TCheckBox_SetLeft(Self As TCheckBox,byval fLeft As Integer)
Declare Sub TCheckBox_SetTop(Self As TCheckBox,byval fTop As Integer)
Declare Sub TCheckBox_SetWidth(Self As TCheckBox,byval fWidth As Integer)
Declare Sub TCheckBox_SetHeight(Self As TCheckBox,byval fHeight As Integer)
Declare Sub TCheckBox_SetVisible(Self As TCheckBox,byval fVisible As Boolean)
Declare Sub TCheckBox_SetParent(Self As TCheckBox,byval fParent As HWND)
Declare Sub TCheckBox_SetEnabled(Self As TCheckBox,byval fEnabled As Boolean)
Declare Sub TCheckBox_SetFont(Self As TCheckBox,fFont As TFont)
Declare Sub TCheckBox_SetPopupMenu(Self As TCheckBox,fPopupMenu As TPopupMenu)

Sub TCheckBox_Initialize(Self As TCheckBox)
    Self.Create       = @TCheckBox_Create
    Self.BringToFront = @TCheckBox_BringToFront
    Self.SetCaption   = @TCheckBox_SetCaption
    Self.SetColor     = @TCheckBox_SetColor
    Self.SetCursor    = @TCheckBox_SetCursor
    Self.SetChecked   = @TCheckBox_SetChecked
    Self.SetFocus     = @TCheckBox_SetFocus
    Self.Free         = @TCheckBox_Free
    Self.SetLeft      = @TCheckBox_SetLeft
    Self.SetTop       = @TCheckBox_SetTop
    Self.SetWidth     = @TCheckBox_SetWidth
    Self.SetHeight    = @TCheckBox_SetHeight
    Self.SetEnabled   = @TCheckBox_SetEnabled
    Self.SetVisible   = @TCheckBox_SetVisible
    Self.SetParent    = @TCheckBox_SetParent
    Self.SetFont      = @TCheckBox_SetFont
    Self.SetPopupMenu = @TCheckBox_SetPopupMenu
    Self.Create(Self)
End Sub

Function TCheckBox_Create(Self As TCheckBox) As HWND
    Self.Params.Style = BS_CHECKBOX
    Self.Params.TypeName = "TCheckBox"
    If Self.Width  = 0 then Self.Width = 75
    If Self.Height = 0 then Self.Height = 17
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"Button",Self.Caption,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TCheckBox_Proc)))
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
        
    If Self.Color = &H0 then Self.Color = GetSysColor(COLOR_BTNFACE)
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    
    Return Self.Handle
End Function

Sub TCheckBox_Free(Self As TCheckBox)
   If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TCheckBox_BringToFront(Self As TCheckBox)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TCheckBox_SetFocus(Self As TCheckBox)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TCheckBox_SetColor(Self As TCheckBox,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle then InvalidateRect(Self.Handle,0,True)
End Sub

Sub TCheckBox_SetCursor(Self As TCheckBox,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TCheckBox_SetChecked(Self As TCheckBox,byval fChecked As Boolean)
    Self.Checked = fChecked
    If Self.Handle then SendMessage(Self.Handle,BM_SETCHECK,Self.Checked,0)
End Sub

Sub TCheckBox_SetCaption(Self As TCheckBox,byval fCaption As String)
    Self.Caption = fCaption
    If Self.Handle then SetWindowText(Self.Handle,Self.Caption)
End Sub

Sub TCheckBox_SetLeft(Self As TCheckBox,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TCheckBox_SetTop(Self As TCheckBox,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TCheckBox_SetWidth(Self As TCheckBox,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TCheckBox_SetHeight(Self As TCheckBox,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TCheckBox_SetEnabled(Self As TCheckBox,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TCheckBox_SetVisible(Self As TCheckBox,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TCheckBox_SetFont(Self As TCheckBox,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If 
End Sub

Sub TCheckBox_SetPopupMenu(Self As TCheckBox,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TCheckBox_SetParent(Self As TCheckBox,byval fParent As HWND)
    Self.Parent = fParent
    If Self.Handle then SetParent(Self.Handle,Self.Parent)
End Sub

Function TCheckBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TCheckBox Ptr
    Dim As RECT WR   
    Dim As POINT PT   
    Dim As Integer X,Y
    Dim As TSmallPoint SP
    Ctrl = Cast(TCheckBox Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
       Case WM_RBUTTONUP: X = LoWord(lParam)
                          Y = HiWord(lParam)
                          PT.X = X
                          PT.Y = Y
                          If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
       Case WM_LBUTTONUP: If Ctrl->Checked = False then 
                             Ctrl->Checked = True
                          Else
                             Ctrl->Checked = False 
                          End If
                          SendMessage(Ctrl->Handle,BM_SETCHECK,Ctrl->Checked,0)
                          If Ctrl->OnClick <> 0 then
                             TCheckBox_DoClick = Ctrl->OnClick
                             TCheckBox_DoClick(*Ctrl)
                          End If                    
       Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_DESTROY  :  RemoveControlProps(Ctrl->Handle) 
       Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If
    End Select
    If Ctrl->Params.Override then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'# TRadioButton #
'################
Type TRadioButton
     Align        As TAlign
     Caption      As String
     Checked      As Boolean
     ClientRect   As RECT
     Color        As Integer
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     Left         As Integer
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     Top          As Integer
     Visible      As Boolean 
     Width        As Integer
  
     Create       As Function(Self As TRadioButton) As HWND
     Free         As Sub(Self As TRadioButton)
     BringToFront As Sub(Self As TRadioButton)
     SetFocus     As Sub(Self As TRadioButton)
     
     SetAlign     As Sub(Self As TRadioButton,byval fAlign As TAlign)
     SetCaption   As Sub(Self As TRadioButton,byval fCaption As String)
     SetChecked   As Sub(Self As TRadioButton,byval fChecked As Boolean)
     SetColor     As Sub(Self As TRadioButton,byval fColor As Integer)
     SetCursor    As Sub(Self As TRadioButton,byval fCursor As HCURSOR)
     SetEnabled   As Sub(Self As TRadioButton,byval fEnabled As Boolean)
     SetFont      As Sub(Self As TRadioButton,Font As TFont)
     SetHeight    As Sub(Self As TRadioButton,byval fHeight As Integer)
     SetLeft      As Sub(Self As TRadioButton,byval fLeft As Integer)
     SetParent    As Sub(Self As TRadioButton,byval fParent As HWND)
     SetPopupMenu As Sub(Self As TRadioButton,PopupMenu As TPopupMenu)
     SetTop       As Sub(Self As TRadioButton,byval fTop As Integer)
     SetVisible   As Sub(Self As TRadioButton,byval fVisible As Boolean)
     SetWidth     As Sub(Self As TRadioButton,byval fWidth As Integer)
     
     OnClick      As Sub(Self As TRadioButton)
End Type

Dim Shared TRadioButton_DoClick    As Sub(Self As TRadioButton)

Declare Function TRadioButton_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TRadioButton_Create(Self As TRadioButton) As HWND
Declare Sub TRadioButton_Free(Self As TRadioButton)
Declare Sub TRadioButton_SetFocus(Self As TRadioButton)
Declare Sub TRadioButton_BringToFront(Self As TRadioButton)
Declare Sub TRadioButton_Initialize(Self As TRadioButton)
Declare Sub TRadioButton_SetCaption(Self As TRadioButton,byval fCaption As String)
Declare Sub TRadioButton_SetColor(Self As TRadioButton,byval fColor As Integer)
Declare Sub TRadioButton_SetCursor(Self As TRadioButton,byval fCursor As HCURSOR)
Declare Sub TRadioButton_SetChecked(Self As TRadioButton,byval fChecked As Boolean)
Declare Sub TRadioButton_SetLeft(Self As TRadioButton,byval fLeft As Integer)
Declare Sub TRadioButton_SetTop(Self As TRadioButton,byval fTop As Integer)
Declare Sub TRadioButton_SetWidth(Self As TRadioButton,byval fWidth As Integer)
Declare Sub TRadioButton_SetHeight(Self As TRadioButton,byval fHeight As Integer)
Declare Sub TRadioButton_SetVisible(Self As TRadioButton,byval fVisible As Boolean)
Declare Sub TRadioButton_SetParent(Self As TRadioButton,byval fParent As HWND)
Declare Sub TRadioButton_SetEnabled(Self As TRadioButton,byval fEnabled As Boolean)
Declare Sub TRadioButton_SetFont(Self As TRadioButton,fFont As TFont)
Declare Sub TRadioButton_SetPopupMenu(Self As TRadioButton,fPopupMenu As TPopupMenu)

Sub TRadioButton_Initialize(Self As TRadioButton)
    Self.Create       = @TRadioButton_Create
    Self.BringToFront = @TRadioButton_BringToFront
    Self.SetCaption   = @TRadioButton_SetCaption
    Self.SetColor     = @TRadioButton_SetColor
    Self.SetCursor    = @TRadioButton_SetCursor
    Self.SetChecked   = @TRadioButton_SetChecked
    Self.SetFocus     = @TRadioButton_SetFocus
    Self.Free         = @TRadioButton_Free
    Self.SetLeft      = @TRadioButton_SetLeft
    Self.SetTop       = @TRadioButton_SetTop
    Self.SetWidth     = @TRadioButton_SetWidth
    Self.SetHeight    = @TRadioButton_SetHeight
    Self.SetEnabled   = @TRadioButton_SetEnabled
    Self.SetVisible   = @TRadioButton_SetVisible
    Self.SetParent    = @TRadioButton_SetParent
    Self.SetFont      = @TRadioButton_SetFont
    Self.SetPopupMenu = @TRadioButton_SetPopupMenu
    Self.Create(Self)
End Sub

Function TRadioButton_Create(Self As TRadioButton) As HWND
    Self.Params.Style = BS_RADIOBUTTON
    Self.Params.TypeName = "TRadioButton"
    If Self.Width  = 0 then Self.Width = 75
    If Self.Height = 0 then Self.Height = 17

    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"Button",Self.Caption,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TRadioButton_Proc)))
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    
    If Self.Color = &H0 then Self.Color = GetSysColor(COLOR_BTNFACE)
    Self.SetColor(Self,Self.Color)
    'Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    
    Application.Controls.Add(Application.Controls,Self.Handle)
    Self.SetChecked(Self,Self.Checked)
    
    Return Self.Handle
End Function

Sub TRadioButton_Free(Self As TRadioButton)
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TRadioButton_BringToFront(Self As TRadioButton)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TRadioButton_SetFocus(Self As TRadioButton)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TRadioButton_SetCaption(Self As TRadioButton,byval fCaption As String)
    Self.Caption = fCaption
    SetWindowText(Self.Handle,Self.Caption)
End Sub

Sub TRadioButton_SetChecked(Self As TRadioButton,byval fChecked As Boolean)
    Self.Checked = fChecked
    If Self.Handle then SendMessage(Self.Handle,BM_SETCHECK,Self.Checked,0)
End Sub

Sub TRadioButton_SetColor(Self As TRadioButton,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle then InvalidateRect(Self.Handle,0,True)
End Sub

Sub TRadioButton_SetCursor(Self As TRadioButton,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TRadioButton_SetLeft(Self As TRadioButton,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TRadioButton_SetTop(Self As TRadioButton,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TRadioButton_SetWidth(Self As TRadioButton,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TRadioButton_SetHeight(Self As TRadioButton,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TRadioButton_SetEnabled(Self As TRadioButton,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TRadioButton_SetVisible(Self As TRadioButton,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TRadioButton_SetFont(Self As TRadioButton,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If 
End Sub

Sub TRadioButton_SetPopupMenu(Self As TRadioButton,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TRadioButton_SetParent(Self As TRadioButton,byval fParent As HWND)
    Self.Parent = fParent
    If Self.Handle then SetParent(Self.Handle,Self.Parent)
End Sub

Sub TRadioButton_TurnOff(Self As TRadioButton)
    Dim i As Integer
    Dim As TParams Ptr P
    Dim As TRadioButton Ptr RB
    Dim As TList Ptr List
    List = cAllocate(Len(TList))
    TList_Initialize(*List)
    GetChilds(Self.Parent,*List)
    For i = 0 to List->Count -1
        P = Cast(TParams Ptr,GetProp(List->Items(i),"Params"))
        If Self.Handle <> List->Items(i) then
            If P->Style and not BS_RADIOBUTTON = 0 then
               RB = Cast(TRadioButton Ptr,GetWindowLong(List->Items(i),GWL_USERDATA))
               SendMessage(List->Items(i),BM_SETCHECK, 0, 0)
               RB->Checked = 0
            End If 
        End If
    Next i
    Deallocate(List)
End Sub

Function TRadioButton_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TRadioButton Ptr
    Dim As RECT WR   
    Dim As POINT PT   
    Dim As Integer X,Y
    Dim As TSmallPoint SP
    Ctrl = Cast(TRadioButton Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
       Case WM_RBUTTONUP: X = LoWord(lParam)
                          Y = HiWord(lParam)
                          PT.X = X
                          PT.Y = Y
                          If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
       Case WM_LBUTTONUP:TRadioButton_TurnOff(*Ctrl)
                          Ctrl->Checked = True
                          Ctrl->SetChecked(*Ctrl,Ctrl->Checked)
                          If Ctrl->OnClick <> 0 then
                             TRadioButton_DoClick = Ctrl->OnClick
                             TRadioButton_DoClick(*Ctrl)
                          End If 
       Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_DESTROY  : RemoveControlProps(Ctrl->Handle)
       Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If
    End Select
    If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#  TGroupBox   #
'################
Type TGroupBox
     Caption      As String
     ClientRect   As RECT
     Color        As Integer
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     Left         As Integer
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     Top          As Integer
     Visible      As Boolean 
     Width        As Integer
     
     BringToFront As Sub(Self As TGroupBox)
     Create       As Function(Self As TGroupBox) As HWND
     Free         As Sub(Self As TGroupBox)
     
     SetCaption   As Sub(Self As TGroupBox,byval fCaption As String)
     SetColor     As Sub(Self As TGroupBox,byval fColor As Integer)
     SetLeft      As Sub(Self As TGroupBox,byval fLeft As Integer)
     SetTop       As Sub(Self As TGroupBox,byval fTop As Integer)
     SetWidth     As Sub(Self As TGroupBox,byval fWidth As Integer)
     SetHeight    As Sub(Self As TGroupBox,byval fHeight As Integer)
     SetParent    As Sub(Self As TGroupBox,byval fParent As HWND)
     SetEnabled   As Sub(Self As TGroupBox,byval fEnabled As Boolean)
     SetVisible   As Sub(Self As TGroupBox,byval fVisible As Boolean)
     SetFont      As Sub(Self As TGroupBox,Font As TFont)
     SetPopupMenu As Sub(Self As TGroupBox,PopupMenu As TPopupMenu)
     
     OnClick      As Sub(Self As TGroupBox)
End Type

Dim Shared TGroupBox_DoClick As Sub(Self As TGroupBox)

Declare Function TGroupBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TGroupBox_Create(Self As TGroupBox) As HWND
Declare Sub TGroupBox_Free(Self As TGroupBox)
Declare Sub TGroupBox_BringToFront(Self As TGroupBox)
Declare Sub TGroupBox_Initialize(Self As TGroupBox)
Declare Sub TGroupBox_SetCaption(Self As TGroupBox,byval fCaption As String)
Declare Sub TGroupBox_SetColor(Self As TGroupBox,byval fColor As Integer)
Declare Sub TGroupBox_SetLeft(Self As TGroupBox,byval fLeft As Integer)
Declare Sub TGroupBox_SetTop(Self As TGroupBox,byval fTop As Integer)
Declare Sub TGroupBox_SetWidth(Self As TGroupBox,byval fWidth As Integer)
Declare Sub TGroupBox_SetHeight(Self As TGroupBox,byval fHeight As Integer)
Declare Sub TGroupBox_SetVisible(Self As TGroupBox,byval fVisible As Boolean)
Declare Sub TGroupBox_SetParent(Self As TGroupBox,byval fParent As HWND)
Declare Sub TGroupBox_SetEnabled(Self As TGroupBox,byval fEnabled As Boolean)
Declare Sub TGroupBox_SetFont(Self As TGroupBox,fFont As TFont)
Declare Sub TGroupBox_SetPopupMenu(Self As TGroupBox,fPopupMenu As TPopupMenu)

Sub TGroupBox_Initialize(Self As TGroupBox)
    Self.Create       = @TGroupBox_Create
    Self.BringToFront = @TGroupBox_BringToFront
    Self.SetCaption   = @TGroupBox_SetCaption
    Self.SetColor     = @TGroupBox_SetColor
    Self.Free         = @TGroupBox_Free
    Self.SetLeft      = @TGroupBox_SetLeft
    Self.SetTop       = @TGroupBox_SetTop
    Self.SetWidth     = @TGroupBox_SetWidth
    Self.SetHeight    = @TGroupBox_SetHeight
    Self.SetEnabled   = @TGroupBox_SetEnabled
    Self.SetVisible   = @TGroupBox_SetVisible
    Self.SetParent    = @TGroupBox_SetParent
    Self.SetFont      = @TGroupBox_SetFont
    Self.SetPopupMenu = @TGroupBox_SetPopupMenu
    Self.Create(Self)
End Sub

Function TGroupBox_Create(Self As TGroupBox) As HWND
    Self.Params.TypeName = "TGroupBox"
    Self.Params.ExStyle   = WS_EX_TRANSPARENT
    Self.Params.Style     = BS_GROUPBOX
    If Self.Width  = 0 then Self.Width = 105
    If Self.Height = 0 then Self.Height = 105
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"Button",Self.Caption,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    If Self.Color = &H0 then Self.Color = GetSysColor(COLOR_BTNFACE)
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TGroupBox_Proc)))
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    
    SetProp(Self.Handle,"Params",@Self.Params)
  
    Return Self.Handle
End Function

Sub TGroupBox_Free(Self As TGroupBox)
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TGroupBox_BringToFront(Self As TGroupBox)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TGroupBox_SetCaption(Self As TGroupBox,byval fCaption As String)
    Self.Caption = fCaption
    If Self.Handle then SetWindowText(Self.Handle,Self.Caption)
End Sub

Sub TGroupBox_SetColor(Self As TGroupBox,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle then InvalidateRect(Self.Handle,0,True)
End Sub

Sub TGroupBox_SetLeft(Self As TGroupBox,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TGroupBox_SetTop(Self As TGroupBox,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TGroupBox_SetWidth(Self As TGroupBox,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TGroupBox_SetHeight(Self As TGroupBox,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TGroupBox_SetEnabled(Self As TGroupBox,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TGroupBox_SetVisible(Self As TGroupBox,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TGroupBox_SetFont(Self As TGroupBox,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If 
End Sub

Sub TGroupBox_SetPopupMenu(Self As TGroupBox,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TGroupBox_SetParent(Self As TGroupBox,byval fParent As HWND)
    Self.Parent = fParent
    If Self.Handle then SetParent(Self.Handle,Self.Parent)
End Sub

Function TGroupBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TGroupBox Ptr
    Dim As RECT WR   
    Dim As POINT PT   
    Dim As Integer X,Y,I,BK,FG
    Dim As LPTOOLTIPTEXT TTX
    Dim As NMHDR Ptr NH
    Dim As DRAWITEMSTRUCT Ptr diStruct
    Dim As MEASUREITEMSTRUCT Ptr miStruct
    Dim As HWND hChild
    Dim As HDC DC
    Dim As HBRUSH Brush
    Dim As TParams Ptr Params
    Ctrl = Cast(TGroupBox Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
       Case WM_RBUTTONUP: X = LoWord(lParam)
                          Y = HiWord(lParam)
                          PT.X = X
                          PT.Y = Y
                          If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
       Case WM_LBUTTONUP: If Ctrl->OnClick <> 0 then
                             TGroupBox_DoClick = Ctrl->OnClick
                             TGroupBox_DoClick(*Ctrl)
                          End If 
       Case WM_NOTIFY   :NH = Cast(NMHDR Ptr, lParam)
                          Select Case NH->Code
                                  Case TCN_SELCHANGE   :PostMessage(Cast(HWND,NH->hwndFrom),CM_TABCHANGE,0,0)
                                  Case TCN_SELCHANGING :SendMessage(Cast(HWND,NH->hwndFrom),CM_TABCHANGING,0,0) 
                                  Case TTN_NEEDTEXT    :TTX = Cast(LPTOOLTIPTEXT,lParam)
                                                         For i = 0 to Application.Controls.Count -1
                                                              If GetClass(Application.Controls.Items(i)) = TOOLBARCLASSNAME then
                                                                 If SendMessage(Application.Controls.Items(i),TB_COMMANDTOINDEX,TTX->hdr.idFrom,0) >= 0 then
                                                                    SendMessage(Application.Controls.Items(i),CM_NEEDTEXT,0,lParam)
                                                                    Exit For
                                                                 End If   
                                                              End If
                                                          Next i    
                          End Select
       Case WM_COMMAND  :Select Case HiWord(wParam) And &HFFFF
                                 Case EN_CHANGE :PostMessage(Cast(HWND,LParam),CM_EDITCHANGE,0,0)
                          End Select  
       Case WM_DRAWITEM :diStruct = Cast(DRAWITEMSTRUCT Ptr,lParam)
                          Select Case diStruct->CtlType 
                                  Case ODT_LISTBOX  ,_
                                       ODT_COMBOBOX ,_
                                       ODT_LISTVIEW ,_
                                       ODT_STATIC   ,_ 
                                       ODT_TAB      : SendMessage(diStruct->hwndItem,CM_DRAWITEM,0,lParam)
                          End Select
                          Exit Function
       Case WM_MEASUREITEM :miStruct = Cast(MEASUREITEMSTRUCT Ptr,lParam)
                             Select Case miStruct->CtlType 
                                     Case ODT_LISTBOX ,_ 
                                          ODT_COMBOBOX: SendDlgItemMessage(hWnd,wParam,CM_MEASUREITEM,0,lParam)
                             End Select
                             Exit Function                    
       Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_DESTROY  :  RemoveControlProps(Ctrl->Handle)
       Case WM_CTLCOLORMSGBOX To WM_CTLCOLORSTATIC:
                            If Msg = WM_CTLCOLORBTN and LoByte(GetVersion) >= 5 then Exit Function
                            hChild = Cast(HWND,lParam)
                            DC = Cast(HDC,WParam)
                            FG = Cast(Integer,GetProp(hChild,"FG"))
                            BK = Cast(Integer,GetProp(hChild,"BK")) 
                            SetBKMode(DC,TRANSPARENT)
                            SetBKColOR(DC,BK)
                            SetTextColOR(DC,FG)
                            SetBKMode(DC,OPAQUE)
                            Brush =  CreateSolidBrush(BK)
                            Return cint(Brush)
                            Brushs.Add(Brushs,Brush)
                            Exit Function 
    End Select
    If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#     TEdit    #
'################

Type TEdit
     Align        As TAlign
     CharCase     As TCharCase
     ClientRect   As RECT
     Color        As Integer
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     Hint         As String
     Left         As Integer
     Override     As WNDPROC
     Params       As TParams
     Parent       As HWND
     PasswordChar As String
     PopupMenu    As TPopupMenu
     ReadOnly     As Boolean
     SelLength    As Integer
     SelStart     As Integer
     SelText      As String
     ShowHint     As Boolean
     Text         As String
     Tag          As Integer
     Top          As Integer
     Visible      As Boolean
     Width        As Integer
     
     BringToFront       As Sub(Self As TEdit)
     Clear              As Sub(Self As TEdit)
     ClearSelection     As Sub(Self As TEdit)
     ClearUndo          As Sub(Self As TEdit)
     CopyToClipboard    As Sub(Self As TEdit)
     Create             As Function(Self As TEdit) As HWND
     CutToClipboard     As Sub(Self As TEdit)
     Free               As Sub(Self As TEdit)
     PasteFromClipboard As Sub(Self As TEdit)
     SelectAll          As Sub(Self As TEdit)
     SetFocus           As Sub(Self As TEdit)
     Undo               As Sub(Self As TEdit)
     
     SetAlign        As Sub(Self As TEdit,byval fAlign As TAlign)
     SetCharCase     As Sub(Self As TEdit,byval fCharCase As TCharCase)
     SetColor        As Sub(Self As TEdit,byval fColor As Integer)
     SetCursor       As Sub(Self As TEdit,byval fCursor As HCURSOR)
     SetEnabled      As Sub(Self As TEdit,byval fEnabled As Boolean)
     SetFont         As Sub(Self As TEdit,fFont As TFont)
     SetHeight       As Sub(Self As TEdit,byval fHeight As Integer)
     SetHint         As Sub(Self As TEdit,byval fHint As String)
     SetLeft         As Sub(Self As TEdit,byval fLeft As Integer)
     SetParent       As Sub(Self As TEdit,byval fParent As HWND)
     SetPasswordChar As Sub(Self As TEdit,byval fPasswordChar As String)
     SetPopupMenu    As Sub(Self As TEdit,PopupMenu As TPopupMenu)
     SetShowHint     As Sub(Self As TEdit,byval fShowHint As Boolean)
     SetReadOnly     As Sub(Self As TEdit,byval fReadOnly As Boolean)
     SetSelLength    As Sub(Self As TEdit,byval fSelLength As Integer)
     SetSelStart     As Sub(Self As TEdit,byval fSelStart As Integer)
     SetSelText      As Sub(Self As TEdit,byval fText As String)
     SetText         As Sub(Self As TEdit,byval fText As String)
     SetTop          As Sub(Self As TEdit,byval fTop As Integer)
     SetVisible      As Sub(Self As TEdit,byval fVisible As Boolean)
     SetWidth        As Sub(Self As TEdit,byval fWidth As Integer)
 
     OnChange        As Sub(Self As TEdit)
     OnKeyPress      As Sub(Self As TEdit,byval Key As Byte)
End Type

Dim Shared TEdit_DoChange As Sub(Self As TEdit)
Dim Shared TEdit_DoKeyPress As Sub(Self As TEdit,byval Key As Byte)

Declare Function TEdit_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TEdit_Create(Self As TEdit) As HWND
Declare Sub TEdit_Free(Self As TEdit)
Declare Sub TEdit_SetFocus(Self As TEdit)
Declare Sub TEdit_BringToFront(Self As TEdit)
Declare Sub TEdit_Initialize(Self As TEdit)
Declare Sub TEdit_SetAlign(Self As TEdit,byval fAlign As TAlign)
Declare Sub TEdit_SetText(Self As TEdit,byval fText As String)
Declare Sub TEdit_SetHint(Self As TEdit,byval fHint As String)
Declare Sub TEdit_SetShowHint(Self As TEdit,byval fShowHint As Boolean)
Declare Sub TEdit_SetCursor(Self As TEdit,byval fCursor As HCURSOR)
Declare Sub TEdit_SetLeft(Self As TEdit,byval fLeft As Integer)
Declare Sub TEdit_SetTop(Self As TEdit,byval fTop As Integer)
Declare Sub TEdit_SetWidth(Self As TEdit,byval fWidth As Integer)
Declare Sub TEdit_SetHeight(Self As TEdit,byval fHeight As Integer)
Declare Sub TEdit_SetVisible(Self As TEdit,byval fVisible As Boolean)
Declare Sub TEdit_SetParent(Self As TEdit,byval fParent As HWND)
Declare Sub TEdit_SetEnabled(Self As TEdit,byval fEnabled As Boolean)
Declare Sub TEdit_SetFont(Self As TEdit,fFont As TFont)
Declare Sub TEdit_SetPopupMenu(Self As TEdit,fPopupMenu As TPopupMenu)
Declare Sub TEdit_SetPasswordChar(Self As TEdit,byval fPasswordChar As String)
Declare Sub TEdit_SetCharCase(Self As TEdit,byval fCharCase As TCharCase)
Declare Sub TEdit_Clear(Self As TEdit)
Declare Sub TEdit_SetSelText(Self As TEdit,byval fText As String)
Declare Sub TEdit_SetColor(Self As TEdit,byval fColor As Integer)
Declare Sub TEdit_SetSelLength(Self As TEdit,byval fSelLength As Integer)
Declare Sub TEdit_SetSelStart(Self As TEdit,byval fSelStart As Integer)
Declare Sub TEdit_ClearSelection(Self As TEdit)
Declare Sub TEdit_CopyToClipboard(Self As TEdit)
Declare Sub TEdit_CutToClipboard(Self As TEdit)
Declare Sub TEdit_PasteFromClipboard(Self As TEdit)
Declare Sub TEdit_Undo(Self As TEdit)
Declare Sub TEdit_ClearUndo(Self As TEdit)
Declare Sub TEdit_SelectAll(Self As TEdit)
Declare Sub TEdit_SetReadOnly(Self As TEdit,byval fReadOnly As Boolean)
Declare Sub TEdit_SetSelTextBuf(Self As TEdit,byval fBuffer As String)
     
Sub TEdit_Initialize(Self As TEdit)
    Self.Create             = @TEdit_Create
    Self.BringToFront       = @TEdit_BringToFront
    Self.Clear              = @TEdit_Clear
    Self.Free               = @TEdit_Free
    Self.Undo               = @TEdit_Undo
    Self.CopyToClipboard    = @TEdit_CopyToClipboard
    Self.CutToClipboard     = @TEdit_CutToClipboard
    Self.PasteFromClipboard = @TEdit_PasteFromClipboard
    Self.SelectAll          = @TEdit_SelectAll
    Self.SetFocus           = @TEdit_SetFocus
    Self.SetAlign           = @TEdit_SetAlign
    Self.SetText            = @TEdit_SetText
    Self.SetHint            = @TEdit_SetHint
    Self.SetShowHint        = @TEdit_SetShowHint
    Self.SetCursor          = @TEdit_SetCursor
    Self.SetCharCase        = @TEdit_SetCharCase
    Self.SetLeft            = @TEdit_SetLeft
    Self.SetTop             = @TEdit_SetTop
    Self.SetWidth           = @TEdit_SetWidth
    Self.SetHeight          = @TEdit_SetHeight
    Self.SetEnabled         = @TEdit_SetEnabled
    Self.SetVisible         = @TEdit_SetVisible
    Self.SetParent          = @TEdit_SetParent
    Self.SetFont            = @TEdit_SetFont
    Self.SetPopupMenu       = @TEdit_SetPopupMenu
    Self.SetColor           = @TEdit_SetColor
    Self.SetSelStart        = @TEdit_SetSelStart
    Self.SetSelText         = @TEdit_SetSelText
    Self.SetSelLength       = @TEdit_SetSelLength
    Self.SetVisible         = @TEdit_SetVisible
    Self.SetReadOnly        = @TEdit_SetReadOnly
    Self.SetPasswordChar    = @TEdit_SetPasswordChar
    Self.Create(Self)
End Sub

Function TEdit_Create(Self As TEdit) As HWND
    Self.Params.TypeName = "TEdit"
    Self.Params.Style     = WS_GROUP OR WS_TABSTOP OR ES_LEFT OR ES_AUTOHSCROLL
    Self.Params.ExStyle   = WS_EX_CLIENTEDGE
  
    If Self.Width  = 0 then Self.Width  = 121
    If Self.Height = 0 then Self.Height = 21
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"Edit",Self.Text,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    SetProp(Self.Handle,"Cursor",Self.Cursor)
   
    Self.Params.HintHandle = CreateWindowEx(0,TOOLTIPS_CLASS,"",TTS_ALWAYSTIP, 0,0,0,0,0,NULL,Instance,NULL)

    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SubCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)

    SendMessage(Self.Params.HintHandle,TTM_ADDTOOL, 0,cint(@Self.Params.Ti))
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)

    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TEdit_Proc)))
    
    If Self.Color = &H0 then Self.Color = &HFFFFFF
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    If Self.PasswordChar <> "" then Self.SetPasswordChar(Self,Self.PasswordChar)
   
    SetProp(Self.Handle,"Params",@Self.Params)
    
    Self.SetReadOnly(Self,Self.ReadOnly)
    Self.SetCharCase(Self,Self.CharCase)
    
    Self.SetVisible(Self,0) ' FOR XP
    Self.SetVisible(Self,1)
    
    Return Self.Handle
End Function

Sub TEdit_SetAlign(Self As TEdit,byval fAlign As TAlign)
    Self.Align = fAlign
    SetProp(Self.Handle,"Align",@Self.Align)
    Select Case Self.Align
           Case alNone   : '''do nothing
           Case alLeft   : 
           Case alRight  : 
           Case alTop    : 
           Case alBottom : 
           Case alClient : 
    End Select
End Sub

Sub TEdit_Free(Self As TEdit)
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyWindow(Self.Handle)
    End If   
End Sub

Sub TEdit_SetFocus(Self As TEdit)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TEdit_BringToFront(Self As TEdit)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TEdit_SetCharCase(Self As TEdit,byval fCharCase As TCharCase)
    Dim As Long Style
    Dim As TCharCase PrevCase
    PrevCase = Self.CharCase
    Self.CharCase = fCharCase
    If Self.Handle <> 0 and Self.Visible = True then
       If Self.CharCase <> PrevCase then 
           Style = GetWindowLong(Self.Handle,GWL_STYLE)
           SetWindowLong(Self.Handle,GWL_STYLE,Style and (not PrevCase) Or Self.CharCase)
       End If
    End If   
End Sub

Sub TEdit_SetCursor(Self As TEdit,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TEdit_SetHint(Self As TEdit,byval fHint As String)
    Self.Hint = fHint
    Self.Params.Ti.cbSize   = SizeOf(TOOLINFO)
    Self.Params.Ti.uFlags   = TTF_IDISHWND OR TTF_SubCLASS
    Self.Params.Ti.hwnd     = Self.Parent
    Self.Params.Ti.hinst    = Instance
    Self.Params.Ti.uId      = Cast(Integer,Self.Handle)
    Self.Params.Ti.lpszText = StrPtr(Self.Hint)
    SendMessage(Self.Params.HintHandle, TTM_UPDATETIPTEXT, 0,cint(@Self.Params.Ti))
End Sub

Sub TEdit_SetShowHint(Self As TEdit,byval fShowHint As Boolean)
    Self.ShowHint = fShowHint
    SendMessage(Self.Params.HintHandle,TTM_ACTIVATE,Self.ShowHint,0)
End Sub

Sub TEdit_SetText(Self As TEdit,byval fText As String)
    Self.Text = fText
    If Self.Handle then SetWindowText(Self.Handle,Self.Text)
End Sub

Sub TEdit_Clear(Self As TEdit)
    If Self.Handle then Self.SetText(Self,"")
End Sub

Sub TEdit_SetSelText(Self As TEdit,byval fText As String)
    Self.SelText = fText
    If Self.Handle then SendMessage(Self.Handle, EM_REPLACESEL, 0, cint(StrPtr(fText)))
End Sub

Sub TEdit_SetColor(Self As TEdit,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle then InvalidateRect(Self.Handle,NULL,True)
End Sub

Sub TEdit_SetSelLength(Self As TEdit,byval fSelLength As Integer)
    Dim Selection As TSelection
    If Self.Handle = 0 then Exit Sub
    Self.SelLength = fSelLength
    SendMessage(Self.Handle, EM_GETSEL, cint(@Selection.StartPos), cint(@Selection.EndPos))
    Selection.EndPos = Selection.StartPos + fSelLength
    SendMessage(Self.Handle, EM_SETSEL, Selection.StartPos, Selection.EndPos)
    SendMessage(Self.Handle, EM_SCROLLCARET, 0,0)
End Sub

Sub TEdit_SetSelStart(Self As TEdit,byval fSelStart As Integer)
    Self.SelStart = fSelStart
    If Self.Handle then SendMessage(Self.Handle, EM_SETSEL, fSelStart, fSelStart)
End Sub

Sub TEdit_ClearSelection(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, WM_CLEAR, 0, 0)
End Sub

Sub TEdit_CopyToClipboard(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, WM_COPY, 0, 0)
End Sub

Sub TEdit_CutToClipboard(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, WM_CUT, 0, 0)
End Sub

Sub TEdit_PasteFromClipboard(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, WM_PASTE, 0, 0)
End Sub

Sub TEdit_Undo(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, WM_UNDO, 0, 0)
End Sub

Sub TEdit_ClearUndo(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, EM_EMPTYUNDOBUFFER, 0, 0)
End Sub

Sub TEdit_SelectAll(Self As TEdit)
    If Self.Handle then SendMessage(Self.Handle, EM_SETSEL, 0, -1)
End Sub

Sub TEdit_SetReadOnly(Self As TEdit,byval fReadOnly As Boolean)
    Self.ReadOnly = fReadOnly
    If Self.Handle then SendMessage(Self.Handle, EM_SETREADONLY, Self.ReadOnly, 0)
End Sub

Sub TEdit_SetLeft(Self As TEdit,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TEdit_SetTop(Self As TEdit,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TEdit_SetWidth(Self As TEdit,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TEdit_SetHeight(Self As TEdit,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TEdit_SetEnabled(Self As TEdit,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TEdit_SetVisible(Self As TEdit,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TEdit_SetFont(Self As TEdit,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If   
End Sub

Sub TEdit_SetPopupMenu(Self As TEdit,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TEdit_SetParent(Self As TEdit,byval fParent As HWND)
    If fParent <> Self.Parent then
       Self.Parent = fParent
       If Self.Handle then SetParent(Self.Handle,Self.Parent)
    End If
End Sub

Sub TEdit_SetPasswordChar(Self As TEdit,byval fPasswordChar As String)
    Self.PasswordChar = fPasswordChar
    If Self.Handle then SendMessage(Self.Handle,EM_SETPASSWORDCHAR,Asc(Self.PasswordChar),0)
End Sub

Function TEdit_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TEdit Ptr
    Dim As RECT WR   
    Dim As POINT PT   
    Dim As Integer X,Y
    Dim Buff As String
    Dim L    As Integer
    Dim As TSmallPoint SP
    Ctrl = Cast(TEdit Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
       Case WM_RBUTTONUP :X = LoWord(lParam)
                          Y = HiWord(lParam)
                          PT.X = X
                          PT.Y = Y
                          If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
       Case WM_CHAR      :If Ctrl->OnKeyPress <> 0 then
                             TEdit_DoKeyPress = Ctrl->OnKeyPress
                             TEdit_DoKeyPress(*Ctrl,wParam)
                          End If
       Case WM_KEYDOWN  : 
       Case WM_KEYUP    : 
       Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                          GetWindowRect(Ctrl->Handle,@WR)
                          Ctrl->Width  = WR.Right - WR.Left
                          Ctrl->Height = WR.Bottom - WR.Top
       Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                          PT.X = WR.Left
                          PT.Y = WR.Top
                          ScreenToClient(GetParent(Ctrl->Handle),@PT)
                          Ctrl->Left = PT.X
                          Ctrl->Top  = PT.Y
       Case CM_EDITCHANGE: L = GetWindowTextLength(Ctrl->Handle)
                          Buff = Space(L+1)
                          GetWindowText(Ctrl->Handle,Buff,L+1) 
                          Ctrl->Text = Buff
                          SendMessage(Ctrl->Handle,EM_GETSEL,PT.X,PT.Y)
                          Ctrl->SelStart = PT.X
                          Ctrl->SelLength = PT.Y - PT.X
                          Ctrl->SelText = Mid(Ctrl->Text,Ctrl->SelStart,Ctrl->SelLength)
                          If Ctrl->OnChange <> 0 then
                             TEdit_DoChange = Ctrl->OnChange
                             TEdit_DoChange(*Ctrl)
                          End If
       Case WM_MOUSEMOVE:  If GetProp(Ctrl->Handle,"Cursor") then
                              SetCursor(Ctrl->Cursor)
                           End If 
       Case WM_DESTROY  :  RemoveControlProps(Ctrl->Handle)
       Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If
                        
   End Select
   If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
   Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#   TListBox   #
'################

Enum TListBoxStyle
     lbStandard
     lbOwnerDrawFixed
     lbOwnerDrawVariable
End Enum      

Type TListBox
     Align        As TAlign 
     ClientRect   As RECT
     Color        As Integer
     Cursor       As HCURSOR
     Enabled      As Boolean
     Font         As TFont
     Handle       As HWND
     Height       As Integer
     ItemCount    As Integer
     ItemIndex    As Integer
     Left         As Integer
     Params       As TParams
     Parent       As HWND
     PopupMenu    As TPopupMenu
     Style        As TListBoxStyle
     Text         As String
     Top          As Integer
     Visible      As Boolean
     Width        As Integer
     
     AddItem      As Sub(Self As TListBox,byval fItem As String)
     BringToFront As Sub(Self As TListBox)
     Clear        As Sub(Self As TListBox)
     Create       As Function(Self As TListBox) As HWND
     DeleteItem   As Sub(Self As TListBox,byval fIndex As Integer)
     Free         As Sub(Self As TListBox)
     InsertItem   As Sub(Self As TListBox,byval fIndex As Integer,byval fItem As String)
     SetFocus     As Sub(Self As TListBox)
     SetItemData  As Sub(Self As TListBox,byval fIndex As Integer, fData As Any Ptr)
     GetItemData  As Function(Self As TListBox,byval fIndex As Integer) As Any Ptr
     
     SetAlign     As Sub(Self As TListBox,byval fAlign As TAlign)
     SetColor     As Sub(Self As TListBox,byval fColor As Integer)
     SetCursor    As Sub(Self As TListBox,byval fCursor As HCURSOR)
     SetEnabled   As Sub(Self As TListBox,byval fEnabled As Boolean)
     SetFont      As Sub(Self As TListBox,Font As TFont)
     SetHeight    As Sub(Self As TListBox,byval fHeight As Integer)
     SetLeft      As Sub(Self As TListBox,byval fLeft As Integer)
     SetParent    As Sub(Self As TListBox,byval fParent As HWND)
     SetPopupMenu As Sub(Self As TListBox,PopupMenu As TPopupMenu)
     SetStyle     As Sub(Self As TListBox,byval fStyle As TListBoxStyle)
     SetText      As Sub(Self As TListBox,byval fText As String)
     SetTop       As Sub(Self As TListBox,byval fTop As Integer)
     SetVisible   As Sub(Self As TListBox,byval fVisible As Boolean)
     SetWidth     As Sub(Self As TListBox,byval fWidth As Integer)
     
     OnClick       As Sub(Self As TListBox)
     OnDrawItem    As Sub(Self As TListBox,byval ItemIndex As Integer,byval State As Integer, R As Rect)     
     OnMeasureItem As Sub(Self As TListBox,byval ItemIndex As Integer,byval Height As Integer)
End Type

Dim Shared TListBox_DoClick       As Sub(Self As TListBox)
Dim Shared TListBox_DoDrawItem    As Sub(Self As TListBox,byval ItemIndex As Integer,byval State As Integer, R As Rect)
Dim Shared TListBox_DoMeasureItem As Sub(Self As TListBox,byval ItemIndex As Integer,byval Height As Integer)
     
Declare Function TListBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TListBox_Create(Self As TListBox) As HWND
Declare Sub TListBox_SetAlign(Self As TListBox,byval fAlign As TAlign)
Declare Sub TListBox_AddItem(Self As TListBox,byval fItem As String)
Declare Sub TListBox_SetItemData(Self As TListBox,byval fIndex As Integer, fData As Any Ptr)
Declare Function TListBox_GetItemData(Self As TListBox,byval fIndex As Integer) As Any Ptr
Declare Sub TListBox_InsertItem(Self As TListBox,byval fIndex As Integer,byval fItem As String)
Declare Sub TListBox_DeleteItem(Self As TListBox,byval fIndex As Integer)
Declare Sub TListBox_Free(Self As TListBox)
Declare Sub TListBox_Clear(Self As TListBox)
Declare Sub TListBox_SetFocus(Self As TListBox)
Declare Sub TListBox_BringToFront(Self As TListBox)
Declare Sub TListBox_Initialize(Self As TListBox)
Declare Sub TListBox_SetColor(Self As TListBox,byval fColor As Integer)
Declare Sub TListBox_SetCursor(Self As TListBox,byval fCursor As HCURSOR)
Declare Sub TListBox_SetStyle(Self As TListBox,byval fStyle As TListBoxStyle)
Declare Sub TListBox_SetText(Self As TListBox,byval fText As String)
Declare Sub TListBox_SetLeft(Self As TListBox,byval fLeft As Integer)
Declare Sub TListBox_SetTop(Self As TListBox,byval fTop As Integer)
Declare Sub TListBox_SetWidth(Self As TListBox,byval fWidth As Integer)
Declare Sub TListBox_SetHeight(Self As TListBox,byval fHeight As Integer)
Declare Sub TListBox_SetVisible(Self As TListBox,byval fVisible As Boolean)
Declare Sub TListBox_SetParent(Self As TListBox,byval fParent As HWND)
Declare Sub TListBox_SetEnabled(Self As TListBox,byval fEnabled As Boolean)
Declare Sub TListBox_SetFont(Self As TListBox,fFont As TFont)
Declare Sub TListBox_SetPopupMenu(Self As TListBox,fPopupMenu As TPopupMenu)

Sub TListBox_Initialize(Self As TListBox)
    Self.SetAlign     = @TListBox_SetAlign
    Self.Create       = @TListBox_Create
    Self.Clear        = @TListBox_Clear
    Self.BringToFront = @TListBox_BringToFront
    Self.SetItemData  = @TListBox_SetItemData
    Self.GetItemData  = @TListBox_GetItemData
    Self.AddItem      = @TListBox_AddItem
    Self.DeleteItem   = @TListBox_DeleteItem
    Self.InsertItem   = @TListBox_InsertItem
    Self.SetText      = @TListBox_SetText
    Self.SetFocus     = @TListBox_SetFocus
    Self.Free         = @TListBox_Free
    Self.SetStyle     = @TListBox_SetStyle
    Self.SetColor     = @TListBox_SetColor
    Self.SetCursor    = @TListBox_SetCursor
    Self.SetLeft      = @TListBox_SetLeft
    Self.SetTop       = @TListBox_SetTop
    Self.SetWidth     = @TListBox_SetWidth
    Self.SetHeight    = @TListBox_SetHeight
    Self.SetEnabled   = @TListBox_SetEnabled
    Self.SetVisible   = @TListBox_SetVisible
    Self.SetParent    = @TListBox_SetParent
    Self.SetFont      = @TListBox_SetFont
    Self.SetPopupMenu = @TListBox_SetPopupMenu
    Self.Create(Self)
End Sub

Function TListBox_Create(Self As TListBox) As HWND
    Self.Params.TypeName = "TListBox"
    Self.Params.ExStyle   = WS_EX_CLIENTEDGE
    Self.Params.Style     = LBS_HASSTRINGS OR LBS_NOTIFY 
    If Self.Width  = 0 then Self.Width  = 105
    If Self.Height = 0 then Self.Height = 105
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"ListBox",Self.Text,WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TListBox_Proc)))
    
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    
    If Self.Color = &H0 then Self.Color = &HFFFFFF
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    
    Self.ItemIndex = -1
    Self.ItemCount = 0
    
    Return Self.Handle
End Function

Sub TListBox_SetAlign(Self As TListBox,byval fAlign As TAlign)
    Self.Align = fAlign
End Sub

Sub TListBox_SetStyle(Self As TListBox,byval fStyle As TListBoxStyle)
    Self.Style = fStyle
End Sub

Sub TListBox_AddItem(Self As TListBox,byval fItem As String)
    Self.ItemCount += 1
    If Self.Handle  then SendMessage(Self.Handle,LB_ADDSTRING,0,cint(StrPtr(fItem)))
End Sub

Sub TListBox_SetItemData(Self As TListBox,byval fIndex As Integer, fData As Any Ptr)
    If fIndex < 0 or fIndex >= Self.ItemCount then
       MessageBox(Application.Handle,"Invalid itemindex. Out of range.","TListBox",MB_ICONERROR or MB_TASKMODAL)
    Else   
       If Self.Handle  then SendMessage(Self.Handle,LB_SETITEMDATA,fIndex,cint(fData))
    End If   
End Sub

Function TListBox_GetItemData(Self As TListBox,byval fIndex As Integer) As Any Ptr
    If fIndex < 0 or fIndex >= Self.ItemCount then
       MessageBox(Application.Handle,"Invalid item index. Out of range.","TListBox",MB_ICONERROR or MB_TASKMODAL)
    Else   
       If Self.Handle  then Return Cast(Any Ptr,SendMessage(Self.Handle,LB_GETITEMDATA,fIndex,0))
    End If   
End Function

Sub TListBox_InsertItem(Self As TListBox,byval fIndex As Integer,byval fItem As String)
    Self.ItemCount += 1
    If Self.Handle  then SendMessage(Self.Handle,LB_INSERTSTRING,fIndex,cint(StrPtr(fItem)))
End Sub

Sub TListBox_DeleteItem(Self As TListBox,byval fIndex As Integer)
    Self.ItemCount -= 1
    If Self.ItemCount = 0 then Self.ItemIndex = -1
    If Self.Handle  then 
       SendMessage(Self.Handle,LB_SETITEMDATA,fIndex,NULL)
       SendMessage(Self.Handle,LB_DELETESTRING,fIndex,0)
    End If   
End Sub

Sub TListBox_Free(Self As TListBox)
    Self.ItemCount = 0
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TListBox_Clear(Self As TListBox)
    If Self.Handle  then 
       SendMessage(Self.Handle,LB_RESETCONTENT,0,0)
       Self.ItemCount = 0
       Self.ItemIndex = -1
       Self.Text = "" 
    End If   
End Sub

Sub TListBox_BringToFront(Self As TListBox)
    If Self.Handle  then BringWindowToTop(Self.Handle)
End Sub

Sub TListBox_SetFocus(Self As TListBox)
    If Self.Handle  then SetFocus(Self.Handle)
End Sub

Sub TListBox_SetColor(Self As TListBox,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle  then InvalidateRect(Self.Handle,NULL,True)
End Sub

Sub TListBox_SetCursor(Self As TListBox,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TListBox_SetText(Self As TListBox,byval fText As String)
    Dim i As Integer
    Dim s As String,Tk As String,t As String
    If Self.Handle = 0 then Exit Sub
    Self.Clear(Self)
    t = ReplaceSubStr(fText,CRLF,"|",1) + "|"
    For i = 1 to Len(t)
        s = s + Mid(t,i,1)
        If Mid(t,i,1) = "|" then
           Tk = Mid(s,1,Len(s)-1)
           Self.AddItem(Self,Tk)
           s = ""
        End If   
    Next i
    Self.Text = fText
End Sub

Sub TListBox_SetLeft(Self As TListBox,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle  then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TListBox_SetTop(Self As TListBox,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle  then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TListBox_SetWidth(Self As TListBox,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle  then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TListBox_SetHeight(Self As TListBox,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle  then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TListBox_SetEnabled(Self As TListBox,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle  then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TListBox_SetVisible(Self As TListBox,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle  then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TListBox_SetFont(Self As TListBox,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
    End If 
End Sub

Sub TListBox_SetPopupMenu(Self As TListBox,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TListBox_SetParent(Self As TListBox,byval fParent As HWND)
    Self.Parent = fParent
    If Self.Handle  then SetParent(Self.Handle,Self.Parent)
End Sub

Function TListBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
   Dim Ctrl As TListBox Ptr
   Dim As RECT WR   
   Dim As POINT PT   
   Dim As Integer X,Y
   Dim As TSmallPoint SP
   Ctrl = Cast(TListBox Ptr,GetWindowLong(hWnd,GWL_USERDATA))
   Select Case Msg
          Case WM_RBUTTONUP: X = LoWord(lParam)
                             Y = HiWord(lParam)
                             PT.X = X
                             PT.Y = Y
                             If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
          Case WM_LBUTTONUP: Ctrl->ItemIndex = SendMessage(Ctrl->Handle,LB_GETCURSEL,0,0)
                             If Ctrl->OnClick <> 0 then
                                TListBox_DoClick = Ctrl->OnClick
                                TListBox_DoClick(*Ctrl)
                             End If 
          Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                             GetWindowRect(Ctrl->Handle,@WR)
                             Ctrl->Width  = WR.Right - WR.Left
                             Ctrl->Height = WR.Bottom - WR.Top
          Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                             GetWindowRect(Ctrl->Handle,@WR)
                             Ctrl->Width  = WR.Right - WR.Left
                             Ctrl->Height = WR.Bottom - WR.Top
          Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                             PT.X = WR.Left
                             PT.Y = WR.Top
                             ScreenToClient(GetParent(Ctrl->Handle),@PT)
                             Ctrl->Left = PT.X
                             Ctrl->Top  = PT.Y
          Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                             PT.X = WR.Left
                             PT.Y = WR.Top
                             ScreenToClient(GetParent(Ctrl->Handle),@PT)
                             Ctrl->Left = PT.X
                             Ctrl->Top  = PT.Y
                             
          Case WM_DESTROY  :  RemoveControlProps(Ctrl->Handle)
          Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If                   
                   
    End Select
    If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#   TComboBox  #
'################
Enum TComboBoxStyle
    csSimple             = CBS_SIMPLE
    csDropDown           = CBS_DROPDOWN
    csDropDownList       = CBS_DROPDOWNLIST
    csOwnerDrawFixed     = CBS_DROPDOWNLIST or CBS_OWNERDRAWFIXED
    csOwnerDrawVariable  = CBS_DROPDOWNLIST or CBS_OWNERDRAWVARIABLE
End Enum    
Type TComboBox
     Align         As TAlign
     ClientRect    As RECT
     Canvas        As TCanvas
     Color         As Integer
     Cursor        As HCURSOR 
     DropDownCount As Integer
     Enabled       As Boolean
     Font          As TFont
     Handle        As HWND
     Height        As Integer
     ItemCount     As Integer
     ItemHeight    As Integer
     ItemIndex     As Integer
     Left          As Integer
     Params        As TParams
     Parent        As HWND
     PopupMenu     As TPopupMenu
     Style         As TComboBoxStyle
     Text          As String
     Top           As Integer
     Visible       As Boolean
     Width         As Integer
     
     AddItem       As Sub(Self As TComboBox,byval fItem As String)
     BringToFront  As Sub(Self As TComboBox)
     Clear         As Sub(Self As TComboBox)
     Create        As Function(Self As TComboBox) As HWND
     DeleteItem    As Sub(Self As TComboBox,byval fIndex As Integer)
     Free          As Sub(Self As TComboBox)
     InsertItem    As Sub(Self As TComboBox,byval fIndex As Integer,byval fItem As String)
     SetFocus      As Sub(Self As TComboBox)
     
     SetAlign      As Sub(Self As TComboBox,byval fAlign As TAlign)
     SetText       As Sub(Self As TComboBox,byval fText As String)
     SetColor      As Sub(Self As TComboBox,byval fColor As Integer)
     SetCursor     As Sub(Self As TComboBox,byval fCursor As HCURSOR)
     SetItemHeight As Sub(Self As TComboBox,byval fItemHeight As Integer)
     SetLeft       As Sub(Self As TComboBox,byval fLeft As Integer)
     SetTop        As Sub(Self As TComboBox,byval fTop As Integer)
     SetWidth      As Sub(Self As TComboBox,byval fWidth As Integer)
     SetHeight     As Sub(Self As TComboBox,byval fHeight As Integer)
     SetParent     As Sub(Self As TComboBox,byval fParent As HWND)
     SetStyle      As Sub(Self As TComboBox,byval fStyle As TComboBoxStyle)
     SetEnabled    As Sub(Self As TComboBox,byval fEnabled As Boolean)
     SetVisible    As Sub(Self As TComboBox,byval fVisible As Boolean)
     SetFont       As Sub(Self As TComboBox,Font As TFont)
     SetPopupMenu  As Sub(Self As TComboBox,PopupMenu As TPopupMenu)
     
     OnClick       As Sub(Self As TComboBox)
     OnChange      As Sub(Self As TComboBox)
     OnDropDown    As Sub(Self As TComboBox)
     OnDrawItem    As Sub(Self As TComboBox,byval ItemIndex As Integer,byval State As Integer, R As Rect)     
     OnMeasureItem As Sub(Self As TComboBox,byval ItemIndex As Integer,byref Height As Integer)
End Type

Dim Shared TComboBox_DoClick       As Sub(Self As TComboBox)
Dim Shared TComboBox_DoChange      As Sub(Self As TComboBox)
Dim Shared TComboBox_DoDropDown    As Sub(Self As TComboBox)
Dim Shared TComboBox_DoDrawItem    As Sub(Self As TComboBox,byval ItemIndex As Integer,byval State As Integer, R As Rect)
Dim Shared TComboBox_DoMeasureItem As Sub(Self As TComboBox,byval ItemIndex As Integer,byref Height As Integer)
     
Declare Function TComboBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TComboBox_Create(Self As TComboBox) As HWND
Declare Sub TComboBox_AddItem(Self As TComboBox,byval fItem As String)
Declare Sub TComboBox_InsertItem(Self As TComboBox,byval fIndex As Integer,byval fItem As String)
Declare Sub TComboBox_DeleteItem(Self As TComboBox,byval fIndex As Integer)
Declare Sub TComboBox_Free(Self As TComboBox)
Declare Sub TComboBox_Clear(Self As TComboBox)
Declare Sub TComboBox_SetFocus(Self As TComboBox)
Declare Sub TComboBox_BringToFront(Self As TComboBox)
Declare Sub TComboBox_Initialize(Self As TComboBox)
Declare Sub TComboBox_SetColor(Self As TComboBox,byval fColor As Integer)
Declare Sub TComboBox_SetCursor(Self As TComboBox,byval fCursor As HCURSOR)
Declare Sub TComboBox_SetItemHeight(Self As TComboBox,byval fItemHeight As Integer)
Declare Sub TComboBox_SetAlign(Self As TComboBox,byval fAlign As TAlign)
Declare Sub TComboBox_SetStyle(Self As TComboBox,byval fStyle As TComboBoxStyle)
Declare Sub TComboBox_SetText(Self As TComboBox,byval fText As String)
Declare Sub TComboBox_SetLeft(Self As TComboBox,byval fLeft As Integer)
Declare Sub TComboBox_SetTop(Self As TComboBox,byval fTop As Integer)
Declare Sub TComboBox_SetWidth(Self As TComboBox,byval fWidth As Integer)
Declare Sub TComboBox_SetHeight(Self As TComboBox,byval fHeight As Integer)
Declare Sub TComboBox_SetVisible(Self As TComboBox,byval fVisible As Boolean)
Declare Sub TComboBox_SetParent(Self As TComboBox,byval fParent As HWND)
Declare Sub TComboBox_SetEnabled(Self As TComboBox,byval fEnabled As Boolean)
Declare Sub TComboBox_SetFont(Self As TComboBox,fFont As TFont)
Declare Sub TComboBox_SetPopupMenu(Self As TComboBox,fPopupMenu As TPopupMenu)

Sub TComboBox_Initialize(Self As TComboBox)
    Self.SetAlign      = @TComboBox_SetAlign
    Self.Create        = @TComboBox_Create
    Self.Clear         = @TComboBox_Clear
    Self.BringToFront  = @TComboBox_BringToFront
    Self.AddItem       = @TComboBox_AddItem
    Self.DeleteItem    = @TComboBox_DeleteItem
    Self.InsertItem    = @TComboBox_InsertItem
    Self.SetText       = @TComboBox_SetText
    Self.SetFocus      = @TComboBox_SetFocus
    Self.Free          = @TComboBox_Free
    Self.SetColor      = @TComboBox_SetColor
    Self.SetCursor     = @TComboBox_SetCursor
    Self.SetStyle      = @TComboBox_SetStyle
    Self.SetItemHeight = @TComboBox_SetItemHeight
    Self.SetLeft       = @TComboBox_SetLeft
    Self.SetTop        = @TComboBox_SetTop
    Self.SetWidth      = @TComboBox_SetWidth
    Self.SetHeight     = @TComboBox_SetHeight
    Self.SetEnabled    = @TComboBox_SetEnabled
    Self.SetVisible    = @TComboBox_SetVisible
    Self.SetParent     = @TComboBox_SetParent
    Self.SetFont       = @TComboBox_SetFont
    Self.SetPopupMenu  = @TComboBox_SetPopupMenu
    Self.Create(Self)
    Self.Canvas.Window = Self.Handle
    TCanvas_Initialize(Self.Canvas)
    Self.Canvas.Font    = Self.Font
End Sub

Function TComboBox_Create(Self As TComboBox) As HWND
    Self.Params.TypeName = "TComboBox"
    Self.Params.ExStyle = WS_EX_CLIENTEDGE
    Self.Params.Style   = WS_VSCROLL OR CBS_AUTOHSCROLL OR CBS_HASSTRINGS OR CBS_NOINTEGRALHEIGHT 
    If Self.Style = 0 then Self.Style = csDropDown
    If Self.DropDownCount = 0 then Self.DropDownCount = 8
    If Self.Width  = 0 then Self.Width  = 121
    If Self.Height = 0 then Self.Height = 21
    
    Self.Handle = CreateWindowEx(Self.Params.ExStyle,"ComboBox","",WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS or Self.Params.Style or Self.Style,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    
    Self.Font.Control = Self.Handle
    TFont_Initialize(Self.Font)
    
    Self.PopupMenu.Window = Self.Handle
    TPopupMenu_Initialize(Self.PopupMenu)
    
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TComboBox_Proc)))
    
    If Self.ItemHeight < 16 then Self.ItemHeight = 16
    Self.SetItemHeight(Self,Self.ItemHeight)
    
    If Self.Color = &H0 then  Self.Color = &HFFFFFF
    Self.SetColor(Self,Self.Color)
    Self.Font.SetColor(Self.Font,Self.Font.Color)
    
    SetProp(Self.Handle,"Params",@Self.Params)
    SetProp(Self.Handle,"Cursor",Self.Cursor)
    
    Return Self.Handle
End Function

Sub TComboBox_SetStyle(Self As TComboBox,byval fStyle As TComboBoxStyle)
    Dim OldIndex As Integer
    Dim OldHandle As HWND
    Dim As Integer i,id
    Dim As String s
    Dim List As TStringList Ptr
    If Self.Handle = 0 then Exit Sub
    List = Allocate(SizeOf(TStringList))
    TStringList_Initialize(*List)
    List->Clear(*List)
    OldHandle = Self.Handle
    OldIndex = SendMessage(Self.Handle,CB_GETCURSEL,0,0)
    For i = 0 to Self.ItemCount -1
        s = Space(SendMessage(Self.Handle,CB_GETLBTEXTLEN,i,0) + 1)
        SendMessage(Self.Handle,CB_GETLBTEXT,i,cint(StrPtr(s)))
        List->Add(*List,s)
        s = ""
    Next i
    id = Application.Controls.IndexOf(Application.Controls,OldHandle)
    Self.Style = fStyle or Self.Params.Style
    Self.Free(Self)
    Self.Create(Self)
    Self.Clear(Self)
    For i = 0 to List->Count -1
        Self.AddItem(Self,List->Items(i))
    Next i
    Self.Style = fStyle
    Application.Controls.Items(id)= Self.Handle
    SendMessage(Self.Handle,CB_SETCURSEL,OldIndex,0)
    Deallocate(List)
End Sub

Sub TComboBox_SetAlign(Self As TComboBox,byval fAlign As TAlign)
    Self.Align = fAlign
End Sub

Sub TComboBox_SetColor(Self As TComboBox,byval fColor As Integer)
    Self.Color = fColor
    SetProp(Self.Handle,"BK",cast(Integer ptr,Self.Color))
    If Self.Handle then InvalidateRect(Self.Handle,NULL,True)
End Sub

Sub TComboBox_SetCursor(Self As TComboBox,byval fCursor As HCURSOR)
    Self.Cursor = fCursor
    SetProp(Self.Handle,"Cursor",Self.Cursor)
End Sub

Sub TComboBox_SetItemHeight(Self As TComboBox,byval fItemHeight As Integer)
    Self.ItemHeight = fItemHeight
    If Self.Style <> csOwnerDrawVariable then
       'Self.ItemHeight = 16
       If Self.Handle then 
          SendMessage(Self.Handle,CB_SETITEMHEIGHT,-1,Self.ItemHeight)
          SendMessage(Self.Handle,CB_SETITEMHEIGHT,0,Self.ItemHeight)
       End If   
    Else
       If Self.ItemHeight < 13 then Self.ItemHeight = 13 
    End If   
End Sub

Sub TComboBox_AddItem(Self As TComboBox,byval fItem As String)
    Self.ItemCount += 1
    If Self.Handle then 
       SendMessage(Self.Handle,CB_ADDSTRING,0,cint(StrPtr(fItem)))
       MoveWindow(Self.Handle, Self.Left, Self.Top, Self.Width, Self.Height + Self.ItemCount * Self.ItemHeight, True)
    End If   
End Sub

Sub TComboBox_InsertItem(Self As TComboBox,byval fIndex As Integer,byval fItem As String)
    Self.ItemCount += 1
    If Self.Handle then SendMessage(Self.Handle,CB_INSERTSTRING,fIndex,cint(StrPtr(fItem)))
End Sub

Sub TComboBox_DeleteItem(Self As TComboBox,byval fIndex As Integer)
    Self.ItemCount -= 1
    If Self.Handle then SendMessage(Self.Handle,CB_DELETESTRING,fIndex,0)
End Sub

Sub TComboBox_Free(Self As TComboBox)
    If Self.Handle then 
       Self.ItemCount = 0
       ReleaseDC(Self.Handle,Self.Canvas.DC)
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyMenu(Self.PopupMenu.Handle)
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TComboBox_Clear(Self As TComboBox)
    If Self.Handle then 
       SendMessage(Self.Handle,CB_RESETCONTENT,0,0)
       Self.ItemCount = 0
       Self.Text = "" 
       MoveWindow(Self.Handle, Self.Left, Self.Top, Self.Width, 0, True)
    End If   
End Sub

Sub TComboBox_BringToFront(Self As TComboBox)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TComboBox_SetFocus(Self As TComboBox)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TComboBox_SetText(Self As TComboBox,byval fText As String)
End Sub

Sub TComboBox_SetLeft(Self As TComboBox,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TComboBox_SetTop(Self As TComboBox,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TComboBox_SetWidth(Self As TComboBox,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TComboBox_SetHeight(Self As TComboBox,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TComboBox_SetEnabled(Self As TComboBox,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TComboBox_SetVisible(Self As TComboBox,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,fVisible)
End Sub

Sub TComboBox_SetFont(Self As TComboBox,fFont As TFont)
    Self.Font = fFont
    If Self.Font.Handle = 0 then
       If Self.Handle then Self.Font.Control = Self.Handle
       TFont_Initialize(Self.Font)
       Self.Canvas.Font = Self.Font
    End If 
End Sub

Sub TComboBox_SetPopupMenu(Self As TComboBox,fPopupMenu As TPopupMenu)
    Self.PopupMenu = fPopupMenu
    If Self.PopupMenu.Handle = 0 then
       If Self.Handle then Self.PopupMenu.Window = Self.Handle
       TPopupMenu_Initialize(Self.PopupMenu)
    End If
End Sub

Sub TComboBox_SetParent(Self As TComboBox,byval fParent As HWND)
    If fParent <> Self.Parent then
       Self.Parent = fParent
       If Self.Handle then SetParent(Self.Handle,Self.Parent)
    End If
End Sub

Function TComboBox_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
   Dim Ctrl As TComboBox Ptr
   Dim As RECT WR   
   Dim As POINT PT   
   Dim As Integer X,Y,FG,BK
   Dim As HBRUSH Brush
   Dim As HWND hChild 
   Dim As HDC DC
   Dim As PAINTSTRUCT Ps
   Dim As DRAWITEMSTRUCT Ptr diStruct
   Dim As MEASUREITEMSTRUCT Ptr miStruct
   Dim As TSmallPoint SP
   Ctrl = Cast(TComboBox Ptr,GetWindowLong(hWnd,GWL_USERDATA))
   Select Case Msg

          Case CM_DRAWITEM      : diStruct = Cast(DRAWITEMSTRUCT Ptr,lParam)
                                  diStruct->ItemAction = ODA_DRAWENTIRE or ODA_SELECT 
                                  Ctrl->Canvas.DC = diStruct->hDC
                                  If Ctrl->OnDrawItem <> 0 then
                                     TComboBox_DoDrawItem = Ctrl->OnDrawItem
                                     TComboBox_DoDrawItem(*Ctrl,diStruct->ItemId,diStruct->itemState,diStruct->rcItem)
                                  End If
                                  SetBKMode(Ctrl->Canvas.DC,OPAQUE)
                                  Exit Function
          Case WM_MEASUREITEM   : miStruct = Cast(MEASUREITEMSTRUCT Ptr,lParam)
                                  If Ctrl->OnMeasureItem <> 0 then
                                     TComboBox_DoMeasureItem = Ctrl->OnMeasureItem
                                     TComboBox_DoMeasureItem(*Ctrl,miStruct->ItemId,miStruct->itemHeight)
                                  End If
                                  Exit Function
          Case CM_COMBODROPDOWN :If Ctrl->OnDropDown <> 0 then
                                     TComboBox_DoDropDown = Ctrl->OnDropDown
                                     TComboBox_DoDropDown(*Ctrl)
                                  End If
                                  Exit Function
          Case CM_COMBOSELCHANGE: Ctrl->ItemIndex = SendMessage(Ctrl->Handle,CB_GETCURSEL,0,0)
                                  Ctrl->Text = Space(SendMessage(Ctrl->Handle,CB_GETLBTEXTLEN,Ctrl->ItemIndex,0) + 1)
                                  SendMessage(Ctrl->Handle,CB_GETLBTEXT,Ctrl->ItemIndex,cint(StrPtr(Ctrl->Text)))
                                  If Ctrl->OnClick <> 0 then
                                     TComboBox_DoClick = Ctrl->OnClick
                                     TComboBox_DoClick(*Ctrl)
                                  End If
                                  If Ctrl->OnChange <> 0 then
                                     TComboBox_DoChange = Ctrl->OnChange
                                     TComboBox_DoChange(*Ctrl)
                                  End If
          Case CM_COMBOCHANGE   :If Ctrl->OnChange <> 0 then
                                     TComboBox_DoChange = Ctrl->OnChange
                                     TComboBox_DoChange(*Ctrl)
                                  End If                         
          Case WM_RBUTTONUP     : X = LoWord(lParam)
                                  Y = HiWord(lParam)
                                  PT.X = X
                                  PT.Y = Y
                                  If Ctrl->PopupMenu.Handle then Ctrl->PopupMenu.AutoPopup(Ctrl->PopupMenu,PT)
          Case WM_LBUTTONUP     : 
          Case WM_SIZE          : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                  GetWindowRect(Ctrl->Handle,@WR)
                                  Ctrl->Width  = WR.Right - WR.Left
                                  Ctrl->Height = WR.Bottom - WR.Top
          Case WM_SIZING        : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                                  GetWindowRect(Ctrl->Handle,@WR)
                                  Ctrl->Width  = WR.Right - WR.Left
                                  Ctrl->Height = WR.Bottom - WR.Top
          Case WM_MOVE          : GetWindowRect(Ctrl->Handle,@WR)
                                  PT.X = WR.Left
                                  PT.Y = WR.Top
                                  ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                  Ctrl->Left = PT.X
                                  Ctrl->Top  = PT.Y
          Case WM_MOVING        : GetWindowRect(Ctrl->Handle,@WR)
                                  PT.X = WR.Left
                                  PT.Y = WR.Top
                                  ScreenToClient(GetParent(Ctrl->Handle),@PT)
                                  Ctrl->Left = PT.X
                                  Ctrl->Top  = PT.Y
          Case WM_CTLCOLORMSGBOX To WM_CTLCOLORSTATIC:
                                  hChild = Cast(HWND,lParam)
                                  DC = Cast(HDC,WParam)
                                  FG = Cast(Integer,GetProp(Ctrl->Handle,"FG"))
                                  BK = Cast(Integer,GetProp(Ctrl->Handle,"BK"))
                                  SetBKMode(DC,TRANSPARENT)
                                  SetBKColor(DC,BK)
                                  SetTextColor(DC,FG)
                                  SetBKMode(DC,OPAQUE)
                                  Brush =  CreateSolidBrush(BK)
                                  Return cint(Brush)
                                  Brushes.Add(Brushes,Brush)
                                  Exit Function         
          Case WM_DESTROY  :  RemoveControlProps(Ctrl->Handle)
          Case wM_SETCURSOR  : If CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam) = 0 then
                                   GetCursorPos(@PT)
                                   SP = PointToSmallPoint(PT)
                                   If SendMessage(Ctrl->Handle,WM_NCHITTEST, 0, MakeLong(SP.X,SP.Y)) = HTCLIENT then
                                      If GetProp(Ctrl->Handle,"Cursor") then 
                                         SetCursor(Ctrl->Cursor)
                                         Exit Function
                                      End If   
                                   Else
                                      SetCursor(crArrow)
                                      Exit Function
                                   End If
                                End If         
    End Select
    If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function

'################
'#  TScrollBar  #
'################
Enum TScrollBarKind
     sbHorizontal = SBS_HORZ
     sbVertical   = SBS_VERT
End Enum 
Enum TScrollCode
    scLineUp    = SB_LINEUP
    scLineDown  = SB_LINEDOWN
    scPageUp    = SB_PAGEUP
    scPageDown  = SB_PAGEDOWN
    scPosition  = SB_THUMBPOSITION
    scTrack     = SB_THUMBTRACK
    scTop       = SB_TOP
    scBottom    = SB_BOTTOM
    scEndScroll = SB_ENDSCROLL
End Enum
Type TScrollBar
     ClientRect   As RECT
     Enabled      As Boolean
     Handle       As HWND
     Height       As Integer
     Kind         As TScrollBarKind
     LargeChange  As Integer
     Left         As Integer
     MaxRange     As Integer
     MinRange     As Integer
     PageSize     As Integer
     Params       As TParams
     Parent       As HWND
     Position     As Integer
     rtlFactor    As Integer
     SmallChange  As Integer
     Visible      As Boolean
     TabStop      As Boolean
     Top          As Integer
     Width        As Integer
     
     BringToFront As Sub(Self As TScrollBar)
     Create       As Function(Self As TScrollBar) As HWND
     Free         As Sub(Self As TScrollBar)
     SetFocus     As Sub(Self As TScrollBar)
     
     SetPosition  As Sub(Self As TScrollBar,byval fPosition As Integer)
     SetPageSize  As Sub(Self As TScrollBar,byval fPageSize As Integer)
     SetKind      As Sub(Self As TScrollBar,byval fKind As TScrollBarKind)
     SetMax       As Sub(Self As TScrollBar,byval fMax As Integer)
     SetMin       As Sub(Self As TScrollBar,byval fMin As Integer)
     SetLeft      As Sub(Self As TScrollBar,byval fLeft As Integer)
     SetTop       As Sub(Self As TScrollBar,byval fTop As Integer)
     SetWidth     As Sub(Self As TScrollBar,byval fWidth As Integer)
     SetHeight    As Sub(Self As TScrollBar,byval fHeight As Integer)
     SetParent    As Sub(Self As TScrollBar,byval fParent As HWND)
     SetEnabled   As Sub(Self As TScrollBar,byval fEnabled As Boolean)
     SetVisible   As Sub(Self As TScrollBar,byval fVisible As Boolean)
     
     OnChange      As Sub(Self As TScrollBar)
     OnScroll      As Sub(Self As TScrollBar,byval fPosition as integer)
End Type

Dim Shared TScrollBar_DoChange As Sub(Self As TScrollBar)
Dim Shared TScrollBar_DoScroll As Sub(Self As TScrollBar,byval fPosition as integer)

Declare Function TScrollBar_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
Declare Function TScrollBar_Create(Self As TScrollBar) As HWND
Declare Sub TScrollBar_Free(Self As TScrollBar)
Declare Sub TScrollBar_SetFocus(Self As TScrollBar)
Declare Sub TScrollBar_BringToFront(Self As TScrollBar)
Declare Sub TScrollBar_Initialize(Self As TScrollBar)
Declare Sub TScrollBar_SetPosition(Self As TScrollBar,byval fPosition As Integer)
Declare Sub TScrollBar_SetPageSize(Self As TScrollBar,byval fPageSize As Integer)
Declare Sub TScrollBar_SetKind(Self As TScrollBar,byval fKind As TScrollBarKind)
Declare Sub TScrollBar_SetMax(Self As TScrollBar,byval fMax As Integer)
Declare Sub TScrollBar_SetMin(Self As TScrollBar,byval fMin As Integer)
Declare Sub TScrollBar_SetLeft(Self As TScrollBar,byval fLeft As Integer)
Declare Sub TScrollBar_SetTop(Self As TScrollBar,byval fTop As Integer)
Declare Sub TScrollBar_SetWidth(Self As TScrollBar,byval fWidth As Integer)
Declare Sub TScrollBar_SetHeight(Self As TScrollBar,byval fHeight As Integer)
Declare Sub TScrollBar_SetVisible(Self As TScrollBar,byval fVisible As Boolean)
Declare Sub TScrollBar_SetParent(Self As TScrollBar,byval fParent As HWND)
Declare Sub TScrollBar_SetEnabled(Self As TScrollBar,byval fEnabled As Boolean)

Sub TScrollBar_Initialize(Self As TScrollBar)
    Self.Create       = @TScrollBar_Create
    Self.BringToFront = @TScrollBar_BringToFront
    Self.SetPosition  = @TScrollBar_SetPosition
    Self.SetMax       = @TScrollBar_SetMax
    Self.SetMin       = @TScrollBar_SetMin
    Self.SetKind      = @TScrollBar_SetKind
    Self.SetFocus     = @TScrollBar_SetFocus
    Self.Free         = @TScrollBar_Free
    Self.SetLeft      = @TScrollBar_SetLeft
    Self.SetTop       = @TScrollBar_SetTop
    Self.SetWidth     = @TScrollBar_SetWidth
    Self.SetHeight    = @TScrollBar_SetHeight
    Self.SetEnabled   = @TScrollBar_SetEnabled
    Self.SetVisible   = @TScrollBar_SetVisible
    Self.SetParent    = @TScrollBar_SetParent
    Self.Create(Self)
End Sub

Function TScrollBar_Create(Self As TScrollBar) As HWND
    Dim As SCROLLINFO sInfo
    Self.Params.TypeName = "TScrollBar"
    If Self.Width  = 0 then Self.Width = 121
    If Self.Height = 0 then Self.Height = GetSystemMetrics(SM_CYHSCROLL)
    If Self.MaxRange = 0 then Self.MaxRange = 100
    If Self.SmallChange = 0 then Self.SmallChange = 1
    If Self.LargeChange = 0 then Self.LargeChange = 1
    Self.Params.Style =  WS_TABSTOP
    Self.Handle = CreateWindow("ScrollBar","",WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR Self.Params.Style OR Self.Kind,Self.Left,Self.Top,Self.Width,Self.Height,Self.Parent,0,Instance,NULL)
    Application.Controls.Add(Application.Controls,Self.Handle)
    SetWindowLong(Self.Handle,GWL_USERDATA,cint(@Self))
    Self.Params.OldProc = Cast(WNDPROC,SetWindowLong(Self.Handle,GWL_WNDPROC,cint(@TScrollBar_Proc)))
    SetScrollRange(Self.Handle, SB_CTL, Self.MinRange, Self.MaxRange, False)
    sInfo.cbSize = SizeOf(ScrollInfo)
    sInfo.nPage = Self.PageSize
    sInfo.fMask = SIF_PAGE
    SetScrollInfo(Self.Handle, SB_CTL, @sInfo, False)
    If Self.Kind = sbHorizontal then
       SetScrollPos(Self.Handle, SB_CTL, Self.Position, True)
       Self.rtlFactor = 1
    Else
       SetScrollPos(Self.Handle, SB_CTL, Self.MaxRange - Self.Position, True)
       Self.rtlFactor = -1
    End If
    SetProp(Self.Handle,"BK",cast(Integer ptr,clBtnFace))
    SetProp(Self.Handle,"Params",@Self.Params)
    Return Self.Handle
End Function

Sub TScrollBar_Free(Self As TScrollBar)
    If Self.Handle then
       Application.Controls.Remove(Application.Controls,Application.Controls.IndexOf(Application.Controls,Self.Handle))
       DestroyWindow(Self.Handle)
    End If
End Sub

Sub TScrollBar_BringToFront(Self As TScrollBar)
    If Self.Handle then BringWindowToTop(Self.Handle)
End Sub

Sub TScrollBar_SetFocus(Self As TScrollBar)
    If Self.Handle then SetFocus(Self.Handle)
End Sub

Sub TScrollBar_SetPageSize(Self As TScrollBar,byval fPageSize As Integer)
    Dim sInfo As SCROLLINFO
    If Self.Handle = 0 then Exit Sub
    If (Self.PageSize = fPageSize) or (Self.PageSize > Self.MaxRange) then Exit Sub
    Self.PageSize = fPageSize
    sInfo.cbSize = SizeOf(ScrollInfo)
    sInfo.nPage = fPageSize
    sInfo.fMask = SIF_PAGE
    If Self.Handle then  SetScrollInfo(Self.Handle, SB_CTL, @sInfo, True)
End Sub

Sub TScrollBar_Recreate(Self As TScrollBar)
    Dim As SCROLLINFO ScrollInfo
    If Self.Handle  then
       Self.Free(Self)
       Self.Create(Self)
    End If   
End Sub

Sub TScrollBar_SetKind(Self As TScrollBar,byval fKind As TScrollBarKind)
    Dim OldKind As TScrollBarKind
    Dim As Integer Temp 
    If Self.Handle then
       If fKind = Self.Kind then Exit Sub 
       OldKind = Self.Kind
       Self.Kind = fKind
       If OldKind = sbHorizontal then
          Temp = Self.Height 
          Self.Height = Self.Width
          Self.Width  = Temp
       Else
          Temp = Self.Width
          Self.Width = Self.Height
          Self.Height  = Temp 
       End If
       TScrollBar_Recreate(Self)
    End If
End Sub

Sub TScrollBar_SetScroll(Self As TScrollBar,byval APosition As Integer,byval AMin As Integer,byval AMax As Integer)
    If AMax < AMin then Exit Sub ' exception
    If APosition < AMin then APosition = AMin
    If APosition > AMax then APosition = AMax
    If (Self.MinRange <> AMin) or (Self.MaxRange <> AMax) then
        Self.MinRange = AMin
        Self.MaxRange = AMax
        If Self.Handle then SetScrollRange(Self.Handle, SB_CTL, AMin, AMax, (Self.Position = APosition))
    End If
    If Self.Position <> APosition then
       Self.Position = APosition
       If Self.Handle then
          If Self.Kind = sbHorizontal then
             SetScrollPos(Self.Handle, SB_CTL, Self.Position, True)
          Else
             SetScrollPos(Self.Handle, SB_CTL, Self.MaxRange - Self.Position, True)
          End If
       End If   
       If Self.OnChange <> 0 then
          TScrollBar_DoChange = Self.OnChange
          TScrollBar_DoChange(Self)
       End If
    End If
End Sub

Sub TScrollBar_SetPosition(Self As TScrollBar,byval fPosition As Integer)
    If Self.Handle then TScrollBar_SetScroll(Self,fPosition,Self.MinRange,Self.MaxRange)
End Sub

Sub TScrollBar_SetMax(Self As TScrollBar,byval fMax As Integer)
    If Self.Handle then TScrollBar_SetScroll(Self,Self.Position,Self.MinRange,fMax)
End Sub

Sub TScrollBar_SetMin(Self As TScrollBar,byval fMin As Integer)
    If Self.Handle then TScrollBar_SetScroll(Self,Self.Position,fMin,Self.MaxRange)
End Sub

Sub TScrollBar_SetLeft(Self As TScrollBar,byval fLeft As Integer)
    Self.Left = fLeft
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TScrollBar_SetTop(Self As TScrollBar,byval fTop As Integer)
    Self.Top = fTop
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TScrollBar_SetWidth(Self As TScrollBar,byval fWidth As Integer)
    Self.Width = fWidth
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TScrollBar_SetHeight(Self As TScrollBar,byval fHeight As Integer)
    Self.Height = fHeight
    If Self.Handle then MoveWindow(Self.Handle,Self.Left,Self.Top,Self.Width,Self.Height,True)
End Sub

Sub TScrollBar_SetEnabled(Self As TScrollBar,byval fEnabled As Boolean)
    Self.Enabled = fEnabled
    If Self.Handle then EnableWindow(Self.Handle,Self.Enabled)
End Sub

Sub TScrollBar_SetVisible(Self As TScrollBar,byval fVisible As Boolean)
    Self.Visible = fVisible
    If Self.Handle then ShowWindow(Self.Handle,SW_SHOW)
End Sub

Sub TScrollBar_SetParent(Self As TScrollBar,byval fParent As HWND)
    Self.Parent = fParent
    If Self.Handle then SetParent(Self.Handle,Self.Parent)
End Sub

Sub TScrollBar_GoScroll(byval wParam As WPARAM,byval lParam As LPARAM)
    Dim ScrollPos As Integer
    Dim NewPos As Short
    Dim sInfo As SCROLLINFO
    Dim As TScrollBar Ptr Ctrl
    Ctrl = Cast(TScrollBar Ptr,GetWindowLong(Cast(HWND,lParam),GWL_USERDATA))
    NewPos = Ctrl->Position
    Select Case LoWord(wParam)
      Case scLineUp:
        NewPos -= Ctrl->SmallChange * Ctrl->RTLFactor
      Case scLineDown:
        NewPos += Ctrl->SmallChange * Ctrl->RTLFactor
      Case scPageUp:
        NewPos -= Ctrl->LargeChange * Ctrl->RTLFactor
      Case scPageDown:
        NewPos += Ctrl->LargeChange * Ctrl->RTLFactor
      Case  scPosition: 
          sInfo.cbSize = SizeOf(ScrollInfo)
          sInfo.fMask = SIF_ALL
          GetScrollInfo(Ctrl->Handle, SB_CTL, @sInfo)
          NewPos = sInfo.nTrackPos
          If Ctrl->Kind = sbVertical then NewPos = Ctrl->MaxRange - NewPos
      Case scTop:
        NewPos = Ctrl->MinRange
      Case scBottom:
        NewPos = Ctrl->MaxRange
    End Select
    If NewPos < Ctrl->MinRange then NewPos = Ctrl->MinRange
    If NewPos > Ctrl->MaxRange then NewPos = Ctrl->MaxRange
    ScrollPos = NewPos
    If Ctrl->OnScroll <> 0 then
       TScrollBar_DoScroll = Ctrl->OnScroll
       TScrollBar_DoScroll(*Ctrl,Ctrl->Position)
    End If
    Ctrl->SetPosition(*Ctrl,ScrollPos)
End Sub

Function TScrollBar_Proc(byval hWnd As HWND,byval Msg As uINT,byval wParam As WPARAM,byval lParam As LPARAM) As LRESULT
    Dim Ctrl As TScrollBar Ptr
    Dim As RECT WR   
    Dim As POINT PT   
    Dim As Integer X,Y
    Ctrl = Cast(TScrollBar Ptr,GetWindowLong(hWnd,GWL_USERDATA))
    Select Case Msg
            Case CM_HSCROLL  : TScrollBar_GoScroll(wParam,lParam)
            Case CM_VSCROLL  : TScrollBar_GoScroll(wParam,lParam)
            Case WM_SIZE     : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                               GetWindowRect(Ctrl->Handle,@WR)
                               Ctrl->Width  = WR.Right - WR.Left
                               Ctrl->Height = WR.Bottom - WR.Top
            Case WM_SIZING   : GetClientRect(Ctrl->Handle,@Ctrl->ClientRect)
                               GetWindowRect(Ctrl->Handle,@WR)
                               Ctrl->Width  = WR.Right - WR.Left
                               Ctrl->Height = WR.Bottom - WR.Top
            Case WM_MOVE     : GetWindowRect(Ctrl->Handle,@WR)
                               PT.X = WR.Left
                               PT.Y = WR.Top
                               ScreenToClient(GetParent(Ctrl->Handle),@PT)
                               Ctrl->Left = PT.X
                               Ctrl->Top  = PT.Y
            Case WM_MOVING   : GetWindowRect(Ctrl->Handle,@WR)
                               PT.X = WR.Left
                               PT.Y = WR.Top
                               ScreenToClient(GetParent(Ctrl->Handle),@PT)
                               Ctrl->Left = PT.X
                               Ctrl->Top  = PT.Y
    End Select
    If Ctrl->Params.Override <> 0 then CallWindowProc(Ctrl->Params.Override,hWnd,Msg,wParam,lParam)
    Return CallWindowProc(Ctrl->Params.OldProc,hWnd,Msg,wParam,lParam)
End Function


'#######Global - InputBox#######
Sub MessageDlg_OnPaint(Sender As TForm)
    DrawIconEx(Sender.Canvas.DC,8,8,LoadIcon(NULL,cast(LPCTSTR,Sender.Tag)),32,32,0,0,DI_NORMAL)
End Sub

Function MessageDlg (byval Message As String,byval MsgType As TMessageType,byval MsgButton As Integer,byval HelpContext As Long) As Integer
    Dim As TForm Ptr Form
    Dim As TLabel Ptr Label
    Dim As TButton Ptr Button
    Dim As Integer W,H,Ln,i,Length
    Dim As String s,ptk,tk,Wrd
      
    Form   = cAllocate(Len(TForm))
    Label  = cAllocate(Len(TLabel))
    
    Select Case MsgType
           Case mtError        : Form->Tag = cast(Integer,IDI_HAND)
           Case mtWarning      : Form->Tag = cast(Integer,IDI_EXCLAMATION)
           Case mtConfirmation : Form->Tag = cast(Integer,IDI_QUESTION)
           Case mtInformation  : Form->Tag = cast(Integer,IDI_ASTERISK)
           Case Else           : Form->Tag = 0
    End Select
    
    Form->Caption     = Application.Title
    Form->BorderStyle = fbDialog
    Form->OnPaint = @MessageDlg_OnPaint
    TForm_Initialize(*Form)
    
    Ln = 1
    Wrd = ""
    If InStr(Message,Chr(10)) then
    For i = 0 to Len(Message)
        s = s + Chr(Message[i])
        If Message[i] = 10 or i = Len(Message) then 
           If Message[i] = 10 then Ln += 1
           ptk = tk
           tk = Mid(s,1,Len(s))
           If Len(Tk) > Len(pTk) then
              Length = Len(Tk)
              Wrd = Tk
           Else
              Length = Len(pTk) 
              Wrd = pTk
           End If
           s = ""
        End If   
    Next i
    Else
        Wrd = Message
    End If

    H = Form->Canvas.TextHeight(Form->Canvas,"Wg")*Ln
    W = Form->Canvas.TextWidth(Form->Canvas,Wrd)'*Length
    Form->SetWidth(*Form,W + 80)
    
    Label->Parent  = Form->Handle
    Label->Top     = 10
    Label->Left   = 50
    Label->Width  = W
    Label->Height  = H
    Label->Caption = Message
    TLabel_Initialize(*Label)
    
    Select Case MsgButton
           Case mbOk              : Button = cAllocate(Len(TButton))
                                     Button->Parent  = Form->Handle 
                                     Button->Caption = "&Ok"
                                     Button->Top     = Label->Top + Label->Height + 20
                                     Button->Left   = (Form->Width - 75)/2
                                     Button->ModalResult = mrOk
                                     TButton_Initialize(*Button)
                                     Button->SetFocus(*Button)
           Case mbOkCancel        : If Form->Width < 160 then Form->Width = 180
                                     Button = cAllocate(Len(TButton)*2)
                                     Button[0].Parent  = Form->Handle 
                                     Button[0].Caption = "&Ok"
                                     Button[0].Top     = Label->Top + Label->Height + 20
                                     Button[0].Left    = (Form->Width - 160)/2
                                     Button[0].ModalResult = mrOk
                                     TButton_Initialize(Button[0])
                                     Button[1].Parent  = Form->Handle 
                                     Button[1].Caption = "&Cancel"
                                     Button[1].Top     = Label->Top + Label->Height + 20
                                     Button[1].Left    = Button[0].Left + 80
                                     Button[1].ModalResult = mrCancel
                                     TButton_Initialize(Button[1])
                                     Button[0].SetFocus(Button[0])
           Case mbYesNo           : If Form->Width < 160 then Form->Width = 180
                                     Button = cAllocate(Len(TButton)*2)
                                     Button[0].Parent  = Form->Handle 
                                     Button[0].Caption = "&Yes"
                                     Button[0].Top     = Label->Top + Label->Height + 20
                                     Button[0].Left    = (Form->Width - 160)/2
                                     Button[0].ModalResult = mrYes
                                     TButton_Initialize(Button[0])
                                     Button[1].Parent  = Form->Handle 
                                     Button[1].Caption = "&No"
                                     Button[1].Top     = Label->Top + Label->Height + 20
                                     Button[1].Left    = Button[0].Left + 80
                                     Button[1].ModalResult = mrNo
                                     TButton_Initialize(Button[1])
                                     Button[0].SetFocus(Button[0])
           Case mbYesNoCancel     : If Form->Width < 250 then Form->Width = 270
                                     Button = cAllocate(Len(TButton)*3)
                                     Button[0].Parent  = Form->Handle 
                                     Button[0].Caption = "&Yes"
                                     Button[0].Top     = Label->Top + Label->Height + 20
                                     Button[0].Left    = (Form->Width - 250)/2
                                     Button[0].ModalResult = mrYes
                                     TButton_Initialize(Button[0])
                                     Button[1].Parent  = Form->Handle 
                                     Button[1].Caption = "&No"
                                     Button[1].Top     = Label->Top + Label->Height + 20
                                     Button[1].Left    = Button[0].Left + 80
                                     Button[1].ModalResult = mrNo
                                     TButton_Initialize(Button[1])
                                     Button[2].Parent  = Form->Handle 
                                     Button[2].Caption = "&Cancel"
                                     Button[2].Top     = Label->Top + Label->Height + 20
                                     Button[2].Left    = Button[0].Left + 80 + Button[1].Width + 5
                                     Button[2].ModalResult = mrCancel
                                     TButton_Initialize(Button[2])
                                     Button[0].SetFocus(Button[0])
           Case mbRetryCancel     : If Form->Width < 160 then Form->Width = 180
                                     Button = cAllocate(Len(TButton)*2)
                                     Button[0].Parent  = Form->Handle 
                                     Button[0].Caption = "&Retry"
                                     Button[0].Top     = Label->Top + Label->Height + 20
                                     Button[0].Left    = (Form->Width - 160)/2
                                     Button[0].ModalResult = mrRetry
                                     TButton_Initialize(Button[0])
                                     Button[1].Parent  = Form->Handle 
                                     Button[1].Caption = "&Cancel"
                                     Button[1].Top     = Label->Top + Label->Height + 20
                                     Button[1].Left    = Button[0].Left + 80
                                     Button[1].ModalResult = mrCancel
                                     TButton_Initialize(Button[1])
                                     Button[0].SetFocus(Button[0])
           Case mbAbortRetryIgnore: If Form->Width < 250 then Form->Width = 270
                                     Button = cAllocate(Len(TButton)*3)
                                     Button[0].Parent  = Form->Handle 
                                     Button[0].Caption = "&Abort"
                                     Button[0].Top     = Label->Top + Label->Height + 20
                                     Button[0].Left    = (Form->Width - 250)/2
                                     Button[0].ModalResult = mrAbort
                                     TButton_Initialize(Button[0])
                                     Button[1].Parent  = Form->Handle 
                                     Button[1].Caption = "&Retry"
                                     Button[1].Top     = Label->Top + Label->Height + 20
                                     Button[1].Left    = Button[0].Left + 80
                                     Button[1].ModalResult = mrRetry
                                     TButton_Initialize(Button[1])
                                     Button[2].Parent  = Form->Handle 
                                     Button[2].Caption = "&Ignore"
                                     Button[2].Top     = Label->Top + Label->Height + 20
                                     Button[2].Left    = Button[0].Left + 80 + Button[1].Width + 5
                                     Button[2].ModalResult = mrIgnore
                                     TButton_Initialize(Button[2])
                                     Button[0].SetFocus(Button[0])
    End Select 

    Form->SetHeight(*Form,10 + H + 10 + 75)
    Form->SetPosition (*Form,poCenterScreen)
    
    Form->ShowModal(*Form)
    Return Form->ModalResult
    Form->Free(*Form)
    DeAllocate(Form)
    DeAllocate(Label)
    DeAllocate(Button)
End Function

Sub ShowMessage(byval Message As String)
    Dim As TForm Ptr Form
    Dim As TLabel Ptr Label
    Dim As TButton Ptr Button
    Dim As Integer W,H,Ln,i,Length
    Dim As String s,ptk,tk,Wrd
    
    Form   = cAllocate(Len(TForm))
    Label  = cAllocate(Len(TLabel))
    Button = cAllocate(Len(TButton))
    
    Form->Caption     = Application.Title
    Form->BorderStyle = fbDialog
    TForm_Initialize(*Form)
    
    Ln = 1
    Wrd = ""
    If InStr(Message,Chr(10)) then
    For i = 0 to Len(Message)
        s = s + Chr(Message[i])
        If Message[i] = 10 or i = Len(Message) then 
           If Message[i] = 10 then Ln += 1
           ptk = tk
           tk = Mid(s,1,Len(s))
           If Len(Tk) > Len(pTk) then
              Length = Len(Tk)
              Wrd = Tk
           Else
              Length = Len(pTk) 
              Wrd = pTk
           End If
           s = ""
        End If   
    Next i
    Else
        Wrd = Message
    End If

    H = Form->Canvas.TextHeight(Form->Canvas,"Wg")*Ln
    W = Form->Canvas.TextWidth(Form->Canvas,Wrd)
    Form->SetWidth(*Form,W + 40)
    
    Label->Parent  = Form->Handle
    Label->Top     = 10
    Label->Left   = 20
    Label->Width  = W
    Label->Height  = H
    Label->Caption = Message
    TLabel_Initialize(*Label)
    
    Form->SetHeight(*Form,10 + H + 10 + 75)
    Form->SetPosition (*Form,poCenterScreen)
    
    Button->Parent      = Form->Handle
    Button->Caption     = "&Ok"
    Button->ModalResult = mrOk
    Button->Top         = Label->Top + Label->Height + 20
    Button->Left        = (Form->Width - 75)/2
    TButton_Initialize(*Button)
    
    Button->SetFocus(*Button)
    
    Form->ShowModal(*Form)
    Form->Free(*Form)
    DeAllocate(Form)
    DeAllocate(Label)
    DeAllocate(Button)
End Sub

Function InputBox(byval Prompt As String,byval Caption As String,byval Text As String = "") As String
    Dim As TForm Ptr InputBoxForm
    Dim As TLabel Ptr ibLabel
    Dim As TEdit Ptr ibEdit
    Dim As TButton Ptr OkButton,CancelButton
    
    InputBoxForm = cAllocate(Len(TForm))
    ibLabel      = cAllocate(Len(TLabel))
    ibEdit       = cAllocate(Len(TEdit))
    okButton     = cAllocate(Len(TButton))
    cancelButton = cAllocate(Len(TButton))
    
    InputBoxForm->Width       = 300
    InputBoxForm->Height      = 140
    InputBoxForm->Position    = poCenterScreen
    InputBoxForm->Caption     = Caption
    InputBoxForm->BorderStyle = fbDialog
    TForm_Initialize(*InputBoxForm)
  
    ibLabel->Parent  = InputBoxForm->Handle
    ibLabel->Top     = 10
    ibLabel->Left    = 20
    ibLabel->Caption = Prompt
    TLabel_Initialize(*ibLabel)
    
    ibEdit->Parent  = InputBoxForm->Handle
    ibEdit->Top     = ibLabel->Top + ibLabel->Height + 5
    ibEdit->Width   = InputBoxForm->Width - 40
    ibEdit->Left    = 20
    TEdit_Initialize(*ibEdit)
    ibEdit->SetText(*ibEdit,"")
    
    OkButton->Parent      = InputBoxForm->Handle
    OkButton->Caption     = "&Ok"
    OkButton->ModalResult = mrOk
    OkButton->Top         = ibLabel->Top + ibLabel->Height + ibEdit->Height + 20
    OkButton->Left        = InputBoxForm->Width\2 - 85
    TButton_Initialize(*OkButton)
    
    CancelButton->Parent      = InputBoxForm->Handle
    CancelButton->Caption     = "&Cancel"
    CancelButton->ModalResult = mrCancel
    CancelButton->Top         = ibLabel->Top + ibLabel->Height + ibEdit->Height + 20
    CancelButton->Left        = OkButton->Left + OkButton->Width + 10
    TButton_Initialize(*CancelButton)
    
    ibEdit->SetFocus(*ibEdit)
    InputBoxForm->ShowModal(*InputBoxForm)
    If InputBoxForm->ModalResult = mrOk then
       Return ibEdit->Text
    ElseIf InputBoxForm->ModalResult = mrCancel then
       Return ""
    End If
    InputBoxForm->Free(*InputBoxForm)
    DeAllocate(InputBoxForm)
    DeAllocate(ibLabel)
    DeAllocate(ibEdit)
    DeAllocate(OkButton)
    DeAllocate(CancelButton)
End Function
